-- Euphoria Grammar for Ox

[Syntax]
AbortOnError = Yes
IgnoreCase = No
InitKeyWordChars = A-Za-z
KeyWordChars = A-Za-z0-9_-
CommentStart =
CommentEnd =
SingleComment = --
StringStart = "
StringEnd = "
StringEsc = \
CharStart = '
CharEnd = '
CharEsc = \
HexStart = #

[Grammar]

-- foo( x, y, z )
Command     =   IDENTIFIER '(' ParmList 
                {: return { CallProc, $1, $3 } :}
-- foo = 23
            |   IDENTIFIER '=' Expr 
                {: return { Assign, $1, $3 } :}
-- foo += 23
            |   IDENTIFIER '+=' Expr 
                {: return { Assign, $1, { BinaryOp, ADD, { Variable, $1 }, $3 } } :}
-- foo -= 23
            |   IDENTIFIER '-=' Expr 
                {: return { Assign, $1, { BinaryOp, SUB, { Variable, $1 }, $3 } } :}
-- foo *= 23
            |   IDENTIFIER '*=' Expr 
                {: return { Assign, $1, { BinaryOp, MUL, { Variable, $1 }, $3 } } :}
-- foo /= 23
            |   IDENTIFIER '/=' Expr 
                {: return { Assign, $1, { BinaryOp, DIV, { Variable, $1 }, $3 } } :}
-- foo &= 23
            |   IDENTIFIER '&=' Expr 
                {: return { Assign, $1, { BinaryOp, CONCAT, { Variable, $1 }, $3 } } :}
                
                
-- foo[...] = 23
            |   IDENTIFIER '[' Indexes '=' Expr 
                {: return { SetIndex, SET, $1, $3, $5 } :}
-- foo[...] += 23
            |   IDENTIFIER '[' Indexes '+=' Expr 
                {: return { SetIndex, ADD, $1, $3, $5 } :}
-- foo[...] -= 23
            |   IDENTIFIER '[' Indexes '-=' Expr 
                {: return { SetIndex, SUB, $1, $3, $5 } :}
-- foo[...] *= 23
            |   IDENTIFIER '[' Indexes '*=' Expr 
                {: return { SetIndex, MUL, $1, $3, $5 } :}
-- foo[...] /= 23
            |   IDENTIFIER '[' Indexes '/=' Expr 
                {: return { SetIndex, DIV, $1, $3, $5 } :}
-- foo[...] &= 23
            |   IDENTIFIER '[' Indexes '&=' Expr 
                {: return { SetIndex, CONCAT, $1, $3, $5 } :}
                
-- procedure foo( bar )
            |   'procedure' IDENTIFIER '(' ArgList 
                {:  integer flag
                    flag = globalFlag
                    globalFlag = 0
                    return { DefProc, flag, $2, $4 } :}
-- function foo( bar )
            |   'function' IDENTIFIER '(' ArgList 
                {:  integer flag
                    flag = globalFlag
                    globalFlag = 0
                    return { DefFunc, flag, $2, $4 } :}
-- type foo( bar )
            |   'type' IDENTIFIER '(' ArgList
                {:  integer flag
                    flag = globalFlag
                    globalFlag = 0
                    return { DefType, flag, $2, $4 } :}
-- end procedure
            |   'end' 'procedure' 
                {: return { End, Procedure } :}
-- end function
            |   'end' 'function' 
                {: return { End, Function } :}
-- end type
            |   'end' 'type' 
                {: return { End, Type } :}

-- return expr
            |   'return'
            {:  return { Return } :}
            
-- if expr then
            |   'if' Expr 'then' 
            {:  return  { If, $2 } :}

-- elsif
            |   'elsif' Expr 'then' 
            {:  return  { Elsif, $2 } :}

-- else
            |   'else' 
            {:  return  { Else } :}

-- end if
            |   'end' 'if' 
            {:  return  { End, If } :}

-- for i = expr to expr by expr do
            |   'for' IDENTIFIER '=' Expr 'to' Expr 'by' Expr 'do' 
                {: return { For, $2, $4, $6, $8 } :}
-- for i = expr to expr do
            |   'for' IDENTIFIER '=' Expr 'to' Expr 'do' 
                {: return { For, $2, $4, $6, {NumLiteral,"1"} } :}
-- for i <ERROR>   
            |   'for' IDENTIFIER
            {:  oxParseError( "Expected '=' or 'in', not %s", {oxToken} )
                return { NoOp } :}
-- for i = expr to expr <ERROR>
            |   'for' IDENTIFIER '=' Expr 'to' Expr
            {:  oxParseError( "Expected 'do' or 'by', not %s", {oxToken} )
                return { NoOp } :}
-- end for
            |   'end' 'for'             
            {:  return { End, For } :}
-- while expr do
            |   'while' Expr 'do'
            {:  return { While, $2 } :}
-- end while
            |   'end' 'while' 
            {:  return  { End, While } :}          
-- end <ERROR>
            |   'end'
            {:  oxParseError( "Expected 'if', 'for', 'while' or 'loop', not %s", {oxToken} )
                return { NoOp } :}
-- ? x
            |   '?' Expr 
                {: return { CallProc, "print", {{NumLiteral,"1"}, $2} } :}
-- exit
            |   'exit' 
                {: return { Exit } :}

-- with...
            |   'with' Flag
                {: return { NoOp } :}
            |   'without' Flag
                {: return { NoOp } :}

-- include
            |   'include'
                {:  -- leave it to opInclude to parse the filename
                    return { Include } :}

-- global ...
            |   'global'
                {:  return { GlobalFlag } :}

-- <EOF>
            | '<EOF>'
            {:  return { ModuleEnd } :}
-- <ERROR>
            |
            {:  oxParseError( "Expected a command, not %s", {oxToken} )
                return { NoOp } :}


-- declaring variables
Command     =   TYPE VarList
                {:  sequence name           
                    for i = 1 to length( $2 ) do
                        name = createVar( globalFlag, $1, $2[i] )
                    end for
                    globalFlag = 0
                    return { CreateVar, $1 } & $2   :}

VarList     =   IDENTIFIER
                {:  return {$1}                     :}

            |   IDENTIFIER ',' VarList
                {:  return prepend( $3, $1 )        :}

-- constant FOO = 12
Command     =   'constant' IDENTIFIER $pre '=' Expr ConstTail
                <:  -- name has to be defined, in case any of
                    -- the constants in the tail reference it
                    sequence name
                    name = createVar( globalFlag, "object", $2 ) :>
                {:  -- note: $pre is not counted as a position
                    globalFlag = 0
                    return { CreateConst, $2, $4 } & $5 :}
                           
                           
ConstTail   =   ',' IDENTIFIER $pre '=' Expr ConstTail
                <:  sequence name
                    name = createVar( globalFlag, "object", $2 ) :>
                {:  -- note: $pre is not counted as a position
                    return { $2, $4 } & $5 :}
            |
                {:  return {} :}

-- assignments
Command     =   IDENTIFIER '=' Expr
                {:  return { Assign, $1, $3 }       :}

-- Expressions have the following level of precedence:
--
--      multiplication, division    * /
--      addition, subtraction       + -
--      concatonation               &
--      logical comparison          = != < <= > >=
--      boolean                     and or xor

-- tightest level of binding: the actual expression
Expr5   =   '(' Expr ')'
            {:  return $2                       :}

        |   '-' Expr
            {:  return { Negate, $2 }           :}
        
        |   '+' Expr
            {:  return $2                       :}
        
        |   'not' Expr
            {:  return { Not, $2 }              :}
        
        |   '{' '}'
            {:  return { CreateSeq }            :}
            
        |   '{' SeqList '}'
            {:  return prepend( $2, CreateSeq ) :}
        
        |   '.' NUMBER
            {:  return { NumLiteral, "." & $2 }  :}
            
        |   NUMBER
            {:  return { NumLiteral, $1 }       :}
        
        |   HEX
            {:  return { NumLiteral, $1 }       :}
        
        |   CHAR
            {:  return { NumLiteral, $1 }       :}
            
        |   STRING
            {:  return { StrLiteral, $1 }       :}
        
        |   IDENTIFIER
            {:  return { Variable, $1 }         :}  

        |   IDENTIFIER '[' Indexes
            {:  return { GetIndex, $1, $3 } :}

        |   IDENTIFIER '(' ParmList
            {: return { CallFunc, $1, $3 } :}
            
        |   TYPE '(' ParmList
            {: return { CallFunc, $1, $3 } :}
        
-- <ERROR>        
        |   
            {: oxParseError("%s has not been declared (expected expression)", {oxTokenValue} ) :}

SeqList =   Expr            
            {:  return {$1}                     :}
            
        |   Expr ',' SeqList
            {:  return prepend( $3, $1 )        :}                                             


-- multiplication and division
Expr4   =   Expr5 '*' Expr4
            {:  return { BinaryOp, MUL, $1, $3 }          :}
            
        |   Expr5 '/' Expr4
            {:  return { BinaryOp, DIV, $1, $3 }          :}
            
        |   Expr5
            {:  return $1                       :}

-- addition and subtraction
Expr3   =   Expr4 '+' Expr3
            {:  return { BinaryOp, ADD, $1, $3 }          :}

        |   Expr4 '-' Expr3
            {:  return { BinaryOp, SUB, $1, $3 }          :}
        
        |   Expr4
            {:  return $1                       :}

-- concatonation
Expr2   =   Expr3 '&' Expr2
            {:  return { BinaryOp, CONCAT, $1, $3 }       :}
            
        |   Expr3
            {:  return $1                       :}
               
-- comparison
Expr1   =   Expr2 '=' Expr2
            {:  return { Compare, EQ, $1, $3 }         :}
            
        |   Expr2 '!=' Expr2
            {:  return { Compare, NE, $1, $3 }         :}
            
        |   Expr2 '>' Expr2                       
            {:  return { Compare, GT, $1, $3 }         :}
            
        |   Expr2 '>=' Expr2
            {:  return { Compare, GE, $1, $3 }         :}
            
        |   Expr2 '<' Expr2
            {:  return { Compare, LT, $1, $3 }         :}
            
        |   Expr2 '<=' Expr2
            {:  return { Compare, LE, $1, $3 }         :}
            
        |   Expr2
            {:  return $1                           :}

-- logical
Expr    =   Expr1 'and' Expr
            {:  return { And, $1, $3 }              :}
        
        |   Expr1 'or' Expr
            {:  return { Or, $1, $3 }               :}
            
        |   Expr1 'xor' Expr
            {:  return { Xor, $1, $3 }              :}
            
        |   Expr1
            {:  return $1                           :}

-- indexes
Indexes =   Expr ']' '[' Indexes
            {:  return prepend( $4, $1 ) :}
        |   Expr ']'
            {:  return { $1 } :}
        |   Expr '..' Expr ']'
            {: return {{ Slice, $1, $3 }} :}

ParmList =  ')'
            {:  return {} :}
        |   List ')'
            {:  return $1 :}

-- arglist for a procedure foo(...)
ArgList =   ')'
            {:  return {} :}
        |   TYPE IDENTIFIER ArgTail
            {:  return prepend( $3, {$1,$2} ) :}
        |   
            {:  oxParseError( "Expected an argument name, not %s", {oxToken} ) 
                return {} :}
        
ArgTail =   ')'
            {:  return {} :}
        |   ',' TYPE IDENTIFIER ArgTail
            {:  return prepend( $4, {$2,$3} ) :}
        |
            {:  oxParseError( "Expected ',' or ')', not %s", {oxToken} ) 
                return {} :}

-- non-empty list
List    =   Expr            
            {:  return {$1} :}
        |   Expr ',' List
            {:  return {$1} & $3 :}
            
-- with/without flags
Flag    =   'profile'
            {: return {} :}
        
        |   'profile_time'
            {: return {} :}
            
        |   'trace'
            {: return {} :}
            
        |   'warning'
            {: return {} :}
            
        |   'type_check'
            {: return {} :}
            
        |   NUMBER
            {: return {} :}

