// Euphoria library routines



//## Import to save some typing..
//## Mic, 000227
import java.io.*;


class Eu {

    //## Added some variables for file functionality.
    //## Mic, 000227

    /* Allow the first 100 file-numbers to be used by various devices (PRN, AUX etc.) */
    static final int FIRST_FILE_NUMBER	= 100;

    /* Set the maximum number of files to be used in a program to 64 */
    static final int MAX_FILES		= 64;

    /* Create a table to hold all the RandomAccessFile-objects	*/
    static RandomAccessFile[] _all_files = new RandomAccessFile[MAX_FILES];

    static int _file_number		= 0;		

    //## Added some more variables for various purposes
    //## Mic, 000228-000302
    static Seq _command_line = new Seq();

    /* Create an inputstream (keyboard) reader */
    static BufferedReader inData =  new BufferedReader(
                                    new InputStreamReader(System.in));

    /* Define some machine procedures/functions */ 
    static final int M_SEEK  = 19,
	             M_WHERE = 20,
	             M_DIR   = 22,
	             M_CURRENT_DIR = 23,
	             M_ALLOW_BREAK = 42,
	             M_CHECK_BREAK = 43,
	             M_FLUSH = 60,
	             M_LOCK_FILE = 61,
	             M_UNLOCK_FILE = 62,
	             M_CHDIR = 63;



    static final int IntTableCount = 256;
    static final int IntTableOffset = -1;
    static final int IntTableMax = IntTableCount + IntTableOffset;
    static Object trueFlag, falseFlag;
    static Object intTable[] = new Object[IntTableCount];
    static long startTime = System.currentTimeMillis()/1000;

    // initialized the ints
    static {
        for ( int i = 0; i < Eu.IntTableCount; i++ ) {
            Eu.intTable[i] = new Integer( i+Eu.IntTableOffset );
        }                          
        
        // intialize the flags
        Eu.trueFlag = Eu.intTable[ 1-IntTableOffset ];
        Eu.falseFlag = Eu.intTable[ 0-IntTableOffset ];
    }

    // for unaryOp()
    static final int    RAND        = 100,
                        NOT         = 101,
                        NEGATE      = 102,
                        SQRT        = 103,
                        ARCTAN      = 104,
                        LOG         = 105,
                        SIN         = 106,
                        COS         = 107,
                        TAN         = 108,
                        FLOOR       = 109,
                        NOT_BITS    = 110;
                       
    // for binaryOp()
    static final int    SET         = 200,
                        ADD         = 201,
                        SUB         = 202,
                        MUL         = 203,
                        DIV         = 204,
                        LT          = 205, 
                        GE          = 206,
                        EQ          = 207,
                        NE          = 208,
                        LE          = 209,
                        GT          = 210,
                        AND         = 211,
                        OR          = 212,
                        OR_BITS     = 213,
                        XOR_BITS    = 214,
                        AND_BITS    = 215,
                        FLOOR_DIV   = 216,
                        REMAINDER   = 217,
                        POWER       = 218,
                        XOR         = 219,
                        CONCAT      = 220;

    // error
    static void fatal( String s ) {
        System.out.println( "Error: " + s );
        System.exit( 0 );
    }

    // exit function error
    static Object badExit( String s ) {
        System.out.println( "Error: attempt to exit function " + s +
            "without returning a value" );
        System.exit( 0 );
        return Eu.create( 0 );
    }


//////////////////////////////////////////////////////////////////////////////
// Converting Java datatypes to Euphoria datatypes
//////////////////////////////////////////////////////////////////////////////

    // boolean
    static Object create( boolean flag ) {
        // create as an Int
        if (flag) {
            return Eu.create( 1 );
        } else {
            return Eu.create( 0 );
        }
    }


    // integer
    static Object create( int a ) {
        if ( a >= Eu.IntTableOffset && a < IntTableMax )
            return Eu.intTable[ a-IntTableOffset ];
        else
            return new Integer( a );
    }

    // double
    static Object create( double a ) {
        // prefer integer type
        if ( Math.floor(a) == a ) {
            return Eu.create( (int)a );
        }
        else {
            return (Object) new Double( a );
        }
    }

    // string
    static Object create( String a ) {
        // create a new string sequence, with no refcount
        Seq s = new Seq( a );
        s.ref = 0;
        return s;
    }
                 
    // stringbuffer
    static Object create( StringBuffer sb ) {
        Seq s = new Seq( sb.toString() );
        s.ref = 0;
        return s;
    }
    
    // dummy routines
    static void dummyProc() {}
    static Object dummyFunc() { return Eu.create( 1 ); }

    // shut down
    static void abort( Object returnCode ) {
        System.exit( 0 );
    }


//////////////////////////////////////////////////////////////////////////////
// Converting Euphoria datatypes to Java datatypes
//////////////////////////////////////////////////////////////////////////////

                   
                   
    static double toNumber( Object o ) {
        // convert to numeric value
        if (o instanceof Integer)
            return ((Integer)o).doubleValue();
        else
            return ((Double)o).doubleValue();
    }

    static int toInt( Object o ) {
        // convert to numeric value
        if (o instanceof Integer)
            return ((Integer)o).intValue();
        else
            return ((Double)o).intValue();
    }


    static String toString( Object o ) {
        // convert a sequence to a string
        // remove all characters < ' '
        
        StringBuffer sb = new StringBuffer();
        char c;
        
        if (!(o instanceof Seq)) {
            Eu.fatal( "toString: expected sequence, not number" );
        }

        // reduce casting
        Seq s = (Seq)o;

        // convert        
        for ( int i=0; i < (s).length; i++ ) {
            // get char
            c = charAt( s, i );

            // printable?
            if ( c >= ' ' || c == '\n' ) {
                // append
                sb.append( c );
            }
        }
        
        return sb.toString();
        
    }


    // exit test for a 'for' loop
    static boolean forLoop( double index, double step, double finish ) {
        // return true if index is past end of loop
        if ( step < 0 ) {
            return (finish <= index);
        } else {
            return (finish >= index);
        }
    }




//////////////////////////////////////////////////////////////////////////////
// Euphoria Type Tests: integer(), atom(), sequence()
//////////////////////////////////////////////////////////////////////////////


    // implementation of integer()
    static Object isInteger( Object o ) {
        if (o instanceof Seq) {
            // sequence
            return Eu.create( 0 );
        } else if (o instanceof Integer) {
            // integer
            return Eu.create( 1 );
        } else {    
            // create a double to hold the value
            double d = ((Double)o).doubleValue();
            
            // same?
            if (d == Math.floor(d)) {
                // true
                return Eu.create(1);
            } else {   
                // false
                return Eu.create(0);
            }
        }
    }

    // implementation of atom()
    static Object isAtom( Object o ) {
        if (!(o instanceof Seq)) {
            // true
            return Eu.create( 1 );
        } else {   
            // false
            return Eu.create( 0 );
        }
    }
                   
    // implementation of sequence()
    static Object isSequence( Object o ) {
        if (o instanceof Seq) {
            // true
            return Eu.create( 1 );
        } else {   
            // false
            return Eu.create( 0 );
        }
    }


//////////////////////////////////////////////////////////////////////////////
// Sequence manipulation: repeat(), append(), etc.
//////////////////////////////////////////////////////////////////////////////


    // repeat()
    static Object repeat( Object o, Object count ) {
        // check count type
        if (count instanceof Seq) {
            Eu.fatal( "repetition count must be an atom" );
        }
        
        // get the repetition count
        int reps = (int) Eu.toNumber( count );
         

        // create a new sequence        
        Seq sNew = new Seq( reps );
        
        // fill with data
        for( int i = 0; i < reps; i++ ) {
            sNew.data[i] = o;
        }
        
        // sequence?
        if (o instanceof Seq) {
            // increment refcount
            ((Seq)o).ref += reps;
        }
        
        return (Object) sNew;
    }

    // append
    static Object append( Object a, Object b ) {

        Seq sNew;
        
        // check arg
        if (!(a instanceof Seq) ) {
            Eu.fatal("first argument of append must be a sequence");
        }

        // refcount added element
        if (b instanceof Seq) {
            ((Seq)b).ref += 1;
        }

        // copy
        sNew = ((Seq)a).makeUnique();

        // append
        sNew.addElement( b );

        return (Object)sNew;
    }
                

    // prepend
    static Object prepend( Object a, Object b ) {
        
        Object result;
        Seq sNew;
        
        // check arg
        if (!(a instanceof Seq)) {
            Eu.fatal( "first argument of prepend must be a sequence" );
        }

        // refcount added element
        if (b instanceof Seq) {
            ((Seq)b).ref += 1;
        }

        // create a sequence
        sNew = new Seq();
        
        // add the right side element
        sNew.addElement( b );
        
        // concat the left hand side
        for ( int i=0; i < ((Seq)a).length; i++ ) {
            // get item
            result = ((Seq)a).data[i];

            // add to the data
            sNew.addElement( result );

            // reference count?
            if (result instanceof Seq) {
                ((Seq)result).ref+= 1;
            }
        }

        // cast as an object            
        return (Object) sNew;
    }


    // concat(o1, o2)
    static Object concat( Object a, Object b ) {

        Object result;
        Seq sNew;

        // refcount added element
        if (b instanceof Seq) {
            ((Seq)b).ref += 1;
        }

        // both are sequences?
        if (a instanceof Seq && b instanceof Seq) {

            // clone the first
            sNew = ((Seq)a).makeUnique();
            
            // concat the second
            for ( int i=0; i < ((Seq)b).length; i++ ) {
                // get element
                result = ((Seq)b).data[i];
                
                // add to the data
                sNew.addElement( result );
                
                // reference
                if (result instanceof Seq) {
                    ((Seq)result).ref += 1;
                }
                                
            }
        }
        
        else if (a instanceof Seq) {
            // clone the left side
            sNew = ((Seq)a).makeUnique();
            
            // add an element to the end
            sNew.addElement(b);
        }
                    
        else if (b instanceof Seq) {           
        
            // calculate new length
            int len = ((Seq)b).length;            
            
            // create a new sequence of new size
            sNew = new Seq( len+1 );

            // add left side as element
            sNew.data[0] = a;
            
            // add right side elements
            for ( int i=0; i < len; i++ ) {
                // add to the data
                sNew.data[i+1] = ((Seq)b).data[i];
            }
        }
            
        else {
            // number, number
            sNew = new Seq( 2 );
                
            // add the elements
            sNew.data[0] = a;
            sNew.data[1] = b;
        }
        
        return (Object) sNew;
    }                    


    // length()
    static Object length( Object s ) {
        if (!(s instanceof Seq)) {
            Eu.fatal( "length of an atom is undefined" );
        }
        
        // get length
        return Eu.create( ((Seq)s).length );
    }


//////////////////////////////////////////////////////////////////////////////
// Unary and Binary Operations
//////////////////////////////////////////////////////////////////////////////


    // unaryOp() - rand, not, negate, sqrt, etc.
    static Object unaryOp( int op, Object o ) {
        Object result;
    
        if (o instanceof Seq) {
            // create a new sequence
            Seq sNew = new Seq();
            
            // build the new sequence
            for ( int i=0; i < ((Seq)o).length; i++ ) {
                // perform operation
                result = unaryOp( op, ((Seq)o).data[i] );
                
                // add to new sequence
                sNew.addElement( result );
                
                // refcount new element?
                if (result instanceof Seq) {
                    ((Seq)result).ref += 1;
                }
            }
            return (Object) sNew;
            
        } else {
            // number
            switch( op ) {
                case Eu.RAND:
                    return Eu.create( (int) (Math.random() * Eu.toNumber(o) ) + (int)1 );
                    
                case Eu.NOT:
                    if( Eu.toNumber(o) == 0 )
                        return trueFlag;
                    else
                        return falseFlag;
                                                      
                case Eu.NEGATE:
                    return Eu.create( -Eu.toNumber(o) );

                case Eu.SQRT:        
                    return Eu.create( Math.sqrt( Eu.toNumber(o) ) );
                    
                case Eu.ARCTAN:        
                    return Eu.create( Math.atan( Eu.toNumber(o) ) );

                case Eu.LOG:        
                    return Eu.create( Math.log( Eu.toNumber(o) ) );

                case Eu.SIN:           
                    return Eu.create( Math.sin( Eu.toNumber(o) ) );
                                  
                case Eu.COS:        
                    return Eu.create( Math.cos( Eu.toNumber(o) ) );

                case Eu.TAN:        
                    return Eu.create( Math.tan( Eu.toNumber(o) ) );

                case Eu.FLOOR:        
                    return Eu.create( Math.floor( Eu.toNumber(o) ) );

                case Eu.NOT_BITS:        
                    return Eu.create( ~(int) Eu.toNumber(o) );
                    
                default:
                    Eu.fatal("unaryOp: unknown op " + op );
                    return falseFlag;    
                    
            }
        }
    }
    

    // binaryOp() - +, =, *, /, etc.
    static Object binaryOp( int op, Object a, Object b ) {
    
        Object result;
        
        // determine types
        if (a instanceof Seq ) {
            if (b instanceof Seq) {
                // sequence, sequence
                int aLen, bLen; 
                Seq sNew;
            
                // get lengths
                aLen = ((Seq)a).length;
                bLen = ((Seq)b).length;
                
                // must be equal
                if (aLen != bLen) {
                    Eu.fatal( "sequence lengths are not the same (" 
                        + aLen + " != " + bLen + ")" );
                }   
        
                // create a sequence to hold the data
                sNew = new Seq( aLen );

                // evaluate each item
                for ( int i=0; i < aLen; i++ ) {
                    // calculate
                    result = binaryOp( op, ((Seq)a).data[i], ((Seq)b).data[i] );
                    
                    // assign
                    sNew.data[i] = result;
                    
                    // reference
                    if (result instanceof Seq) {
                        ((Seq)result).ref += 1;
                    }
                    
                }
                return (Object) sNew;
            }
            else {
                // sequence, number
                int aLen;
                Seq sNew;
        
                // get length
                aLen = ((Seq)a).length;
                
                // create a sequence to hold the data
                sNew = new Seq( aLen );
                
                // set length  
                sNew.free -= aLen;
                sNew.length = aLen;
        
                // evaluate each item
                for ( int i=0; i < aLen; i++ ) {
                    // store in sequence
                    result = binaryOp( op, ((Seq)a).data[i], b );
                    sNew.data[i] = result;
                    
                    // reference
                    if (result instanceof Seq) {
                        ((Seq)result).ref += 1;
                    }
                    
                    
                }
        
                return (Object) sNew;
            }
        }
        else if (b instanceof Seq) {
            // number, sequence
            int bLen;
            Seq sNew;
    
            // get length
            bLen = ((Seq)b).length;
            
            // create a sequence to hold the data
            sNew = new Seq( bLen );
            
            // set length  
            sNew.free -= bLen;
            sNew.length = bLen;
    
            // evaluate each item
            for ( int i=0; i < bLen; i++ ) {
                // store in sequence
                result = binaryOp( op, a, ((Seq)b).data[i] );
                sNew.data[i] = result;
                
                // reference
                if (result instanceof Seq) {
                    ((Seq)result).ref += 1;
                }
                
                
            }
    
            return (Object) sNew;
        }
        else {
            // number,number
            switch(op) {
                
                case Eu.ADD :
                    return Eu.create( Eu.toNumber( a ) + Eu.toNumber( b ) );
                
                case Eu.SUB :
                    return Eu.create( Eu.toNumber( a ) - Eu.toNumber( b ) );
                
                case Eu.MUL :
                    return Eu.create( Eu.toNumber( a ) * Eu.toNumber( b ) );
                
                case Eu.DIV :
                    // check for division by zero
                    if (Eu.toNumber( b ) == 0) {
                        Eu.fatal("attempt to divide by zero");
                    }                     
                    
                    return Eu.create( Eu.toNumber( a ) / Eu.toNumber( b ) );

                case Eu.FLOOR_DIV :
                    // check for division by zero
                    if (Eu.toNumber( b ) == 0) {
                        Eu.fatal("attempt to divide by zero");
                    }                     

                    // divide and round down                    
                    return Eu.create( Math.floor( Eu.toNumber( a ) / Eu.toNumber( b ) ) );
    
                case Eu.EQ :                    
                    if (Eu.toNumber( a ) == Eu.toNumber( b ))
                        return trueFlag;
                    else 
                        return falseFlag;
    
                case Eu.NE :
                    if (Eu.toNumber( a ) != Eu.toNumber( b ))
                        return trueFlag;
                    else 
                        return falseFlag;
                        
                case Eu.LT :
                    if (Eu.toNumber( a ) < Eu.toNumber( b ))
                        return trueFlag;
                    else 
                        return falseFlag;
                    
                case Eu.LE :
                    if (Eu.toNumber( a ) <= Eu.toNumber( b ))
                        return trueFlag;
                    else 
                        return falseFlag;
                                  
                case Eu.GT :
                    if (Eu.toNumber( a ) > Eu.toNumber( b ))
                        return trueFlag;
                    else 
                        return falseFlag;
                    
                case Eu.GE :
                    if (Eu.toNumber( a ) >= Eu.toNumber( b ))
                        return trueFlag;
                    else 
                        return falseFlag;
                
                case Eu.AND :
                    if (Eu.toNumber( a ) != 0 && Eu.toNumber( b ) != 0 )
                        return trueFlag;
                    else 
                        return falseFlag;
                
                case Eu.OR :
                    if (Eu.toNumber( a ) != 0 || Eu.toNumber( b ) != 0)
                        return trueFlag;
                    else 
                        return falseFlag;

                case Eu.XOR :
                    if (Eu.toNumber( a ) != 0 ^ Eu.toNumber( b ) != 0)
                        return trueFlag;
                    else 
                        return falseFlag;


                case Eu.OR_BITS :
                    return Eu.create( (int)Eu.toNumber( a ) | 
                                      (int)Eu.toNumber( b ) );
                    
                case Eu.XOR_BITS :
                    return Eu.create( (int)Eu.toNumber( a ) ^ 
                                      (int)Eu.toNumber( b ) );

                case Eu.AND_BITS :
                    return Eu.create( (int)Eu.toNumber( a ) & 
                                      (int)Eu.toNumber( b ) );
                           
                case Eu.REMAINDER :
                    return Eu.create(Eu.toNumber( a ) % Eu.toNumber( b ));
                           
                case Eu.POWER :
                    return Eu.create( Math.pow( Eu.toNumber( a ), 
                                                Eu.toNumber( b ) ) );
                    
    
                default :          
                    Eu.fatal( "internal error in binaryOp" );
                    return falseFlag;
            }
            
        }
    }


//////////////////////////////////////////////////////////////////////////////
// Accessing Sequences
//////////////////////////////////////////////////////////////////////////////



    // getItem()
    static Object getItem( Object source, Object index ) {

        // return an object from a sequence
        // make sure fetch value is in range

        int i, len;

        // source must be sequence
        if (!(source instanceof Seq)) {
            Eu.fatal( "attempt to subscript an atom (reading from it)");
        }
    
        // index must be number
        if (index instanceof Seq) {
            Eu.fatal( "subscript must be an atom (reading an element of a sequence)");
        }
    
        // get index, round down to integer
        i = (int)toNumber(index);            

        // get the sequence length
        len = ((Seq)source).length;
            
        if (i < 1 || i > len ) {
            Eu.fatal( "subscript value of " + i +
                " is out of bounds, reading from a sequence length of " + len );
        }       

        // fetch element        
        return ((Seq)source).data[ i-1 ];

    }

    // setItem()
    static Object setItem( int op, Object s, Object index, Object newVal ) {
    
        // set the value in a sequence

        int i, len;
        Object prior, copy;

        // source must be a sequence
        if (!(s instanceof Seq)) {
            Eu.fatal( "attempt to subscript an atom (assigning to it)" );
        }

        // index must be a number
        if (index instanceof Seq) {
            Eu.fatal("subscript must be an atom (assiging to subscript from sequence)");
        }

        // get index, round to integer
        i = (int) Eu.toNumber( index );
        
        // get size of sequence
        len = ((Seq)s).length;
        
        // check range
        if (i < 1 || i > len) {
            Eu.fatal("subscript value of " + i + " is out of bounds,\n" +
                    " assigning to a sequence of length " + len );
        }

        // only version?
        if(((Seq)s).ref < 2) {
            // no need to copy
            copy = s;
        } else {
            // make it unique
            copy = (Object) ((Seq)s).makeUnique();
        }

        // get the prior value
        prior = ((Seq)copy).data[ i-1 ];        

        // perform op?
        if (op != Eu.SET) {
            if (op == Eu.CONCAT) {
                newVal = concat( prior, newVal );
            } else {                   
                newVal = binaryOp( op, prior, newVal );
            }
        }
        
        // replace the prior value
        ((Seq)copy).data[i-1] = newVal; 
        
        // refcount the item
        if (newVal instanceof Seq) {
            // add to the data
            ((Seq)newVal).ref += 1;
        }
           
        // dereference prior
        if (prior instanceof Seq) {
            ((Seq)prior).deref();
        }
        
        // return
        return copy;

    }


    // return slice from sequence
    static Object getSlice( Object source, Object from, Object upto ) {

        int low, high, len;
        Object result;
    
        // sequence?
        if (!(source instanceof Seq)) {
            Eu.fatal( "can't slice an atom" );
        }

        // lower index not a number?
        if (from instanceof Seq) {
            Eu.fatal( "slice lower index is not an atom" );
        }

        // upper index not a number?
        if (upto instanceof Seq) {
            Eu.fatal( "slice upper index is not an atom" );
        }

        // length of sequence
        len = ((Seq)source).length;
        
        // convert range to integers
        low = (int) Eu.toNumber(from);
        high = (int) Eu.toNumber(upto);
        
        // low range check
        if (low < 1) {
            Eu.fatal( "slice lower index is less than 1 (" + low + ")" );
        }

        // high range check
        if (high > len ) {
            Eu.fatal( "slice ends past end of sequence (" + high +
                " out of bounds, reading from a sequence of length " + len );
        }
        
        // negative length
        if ( (high-low) < 0  && (high-low) != -1 ) {
            Eu.fatal( "slice length is less than 0 (" + (high-low) + ")" );
        }

        
        // non-empty slice?
        if ( (high-low) == -1 ) {
            return new Seq();
        }
        else {
            // create an empty sequence
            Seq sNew = new Seq( high-low+1 );
        
            // slice
            for ( int i=low, j=0; i <= high; i++, j++ ) {
                // get item
                result = ((Seq)source).data[i-1];
                
                // add to the data
                sNew.data[j] = result;
                
                // refcount?
                if (result instanceof Seq) {
                    ((Seq)result).ref += 1;
                }
                
            }    
            return (Object)sNew;
        }        
    }
          
          
    // set slice in sequence
    static Object setSlice( int op, Object dst, Object from, Object upto,
                                Object src ) {

        int low, high, srcLen, dstLen, sliceLen;
        boolean singleItem;
        Object merged, item;
    
        // sequence?
        if (!(dst instanceof Seq)) {
            Eu.fatal( "attempt to slice an atom" );
        }

        // lower index not a number?
        if (from instanceof Seq) {
            Eu.fatal( "slice lower index is not an atom" );
        }

        // upper index not a number?
        if (upto instanceof Seq) {
            Eu.fatal( "slice upper index is not an atom" );
        }
        
        // convert range to integers
        low = (int) Eu.toNumber(from);
        high = (int) Eu.toNumber(upto);

        // size of destination and slice
        dstLen = ((Seq)dst).length;
        sliceLen = (high - low) + 1;
        
        // range checks
        if ( low < 1 ) {
            Eu.fatal( "slice lower index is less than 1 (" + low + ")" );
        }
        else if ( sliceLen < 0  && sliceLen != -1 ) {
            Eu.fatal( "slice length is less than 0 (" + sliceLen + ")" );
        }
        else if ( high > dstLen ) {
            Eu.fatal( "slice ends past end of sequence (" + high + " > "
                        + dstLen + ")" );
        }
        
        // what type is the source?
        singleItem = !(src instanceof Seq);

        if (!singleItem) {            
        
            // length of destination
            srcLen = ((Seq)src).length;
                              
            // range checks
            if (srcLen != sliceLen ) {
                Eu.fatal( "lengths do not match on assignment to slice (" +
                            srcLen + " != " + dstLen + ")" );
            }
        }
        
        // non-empty slice?
        // !!! check this !!!
        if (sliceLen == -1 ) {
            // return empty sequence
            return new Seq();
        }
            
        // create a sequence to store data in
        merged = new Seq(dstLen);

        // zero offset
        low -= 1;
        high -= 1;
        
        // replace elements
        for ( int i = 0; i < dstLen; i++ ) {

            // get from source or destination?
            if (i >= low && i <= high) {
                // get from source
                if (singleItem) {
                    // single item
                    item = src;
                } else {
                    // sequence
                    item = ((Seq)src).data[i];
                }
            } else {
                // get from destination
                item = ((Seq)dst).data[i];
            }

            // complex op?
            if (op != Eu.SET) {
                // perform op
                item = binaryOp( op, item, (Object)((Seq)dst).data[i] );
            }
        
            // place in new sequence
            ((Seq)merged).data[i] = item;

            // reference count element
            if (item instanceof Seq) {
                ((Seq)item).ref += 1;
            }
            
        }
        return merged;
    }
          

//////////////////////////////////////////////////////////////////////////////
// Comparisons: compare, equal, find, match, etc.
//////////////////////////////////////////////////////////////////////////////



    // workhorse for compare()
    static int compare1( Object a, Object b ) {
    
        int     i, result, len1, len2, shortest;
        double  d1, d2;
      
        // both sequences?
        if ((a instanceof Seq ) && (b instanceof Seq)) {
        
            // get the lengths
            len1 = ((Seq)a).length;
            len2 = ((Seq)b).length;
            
            // determine shortest
            if (len1<len2) {
                shortest = len1;
            } else {
                shortest = len2;
            };
    
            // walk length of shortest
            for (i=0; i<shortest; i++) {                       

                // compare elements
                result = compare1( ((Seq)a).data[i], ((Seq)b).data[i] );

                // exit on difference
                if (result!=0) {
                    return result;
                };
            };
            
            // no difference between shortest length.
            // result depends on the longest
            if (len1==len2) {
                return 0;
                }
            else if (len1>len2) {
                return 1;
                }
            else {
                return -1;
            }
        }

        // is the first element a sequence?
        else if (a instanceof Seq) { 
            // it's the largest
            return 1;
            }                  
            
        // is the second element a sequence?
        else if (b instanceof Seq) {  
            // it's the largest
            return -1;
            }                  
            
        else {
            // convert to doubles
            d1 = Eu.toNumber(a);
            d2 = Eu.toNumber(b);

            // compare
            if (d1 == d2) {
                return 0;
                }
            else if (d1 > d2) {
                return 1;
                }
            else {
                return -1;
            }
        }
    }

    // compare()
    static Object compare( Object a, Object b ) {              
        // compare, and wrap result
        return Eu.create( compare1( a, b ) );
    }

    // equal()
    static Object equal( Object a, Object b ) {
        if ( compare1( a, b ) == 0 )
            return trueFlag;
        else
            return falseFlag;
    }


    // find()
    static Object find( Object a, Object b ) {

        // check arg type
        if (!(b instanceof Seq)) {
            Eu.fatal( "Second argument of find() must be a sequence" );
        }

        // iterate through the sequence
        for ( int i=0; i < ((Seq)b).length; i++) {
            // compare
            if ( compare1(a, ((Seq)b).data[i]) == 0 ) {
                // return position
                return Eu.create( i+1 );
            }
        };
        // no match        
        return falseFlag;
    }


//////////////////////////////////////////////////////////////////////////////
// Formatted Printing
//////////////////////////////////////////////////////////////////////////////



    static StringBuffer formatPrint( Seq s ) {
    
        // support routine for formatPrint
        double          val;
        Object          o;
        StringBuffer    sb = new StringBuffer();

        // open brace
        sb.append("{");

        // walk the buffer
        for ( int i=0; i < s.length; i++ ) {
            // get an element of the vector
            o = s.data[i];

            // what's the type?
            if (o instanceof Seq) {   
                // recurse      
                sb.append( formatPrint( (Seq) o ) );
                }
            else {
                // number
                sb.append( o );
            }
        
            // comma?
            if (i < s.length-1) {
                sb.append( "," );
            };
             
        }
        
        // close brace
        sb.append("}");
        
        return sb;
    }
    
    
    // formatSprintf()
    static StringBuffer formatSprintf( Object a, Object b ) {

        // support routine for formatPrint                            
        Seq             format, args;
        Object          arg;
        StringBuffer    sb = new StringBuffer();
        int             argNum = 0;
        int             width = 256;
        int             decimal = -1;
        int             digit;
        char            c;

        // is the format a number?
        if (!(a instanceof Seq)) {
            // convert it to a sequence
            format = new Seq();
            format.addElement( a );
        }
        else {
            // no modifications needed
            format = (Seq)a;
        }
        
        // arg list a number?
        if (!(b instanceof Seq)) {
            // convert to a sequence
            args = new Seq();
            args.addElement( b );
        }
        else {
            // cast to sequence
            args = (Seq)b;
        }
        
        // walk the format string
        for ( int i=0; i < format.length; i++ ) {
        
            // get a char
            c = charAt( format, i );

            // format specification?
            if (c=='%') {          
                // what's the width?
                width = 0;
                while ( 1==1 ) {
                    // get a char   
                    c = charAt( format, ++i );
                    
                    // translate the digit
                    if      (c=='0') { digit = 0; }
                    else if (c=='1') { digit = 1; }
                    else if (c=='2') { digit = 2; }
                    else if (c=='3') { digit = 3; }
                    else if (c=='4') { digit = 4; }
                    else if (c=='5') { digit = 5; }
                    else if (c=='6') { digit = 6; }
                    else if (c=='7') { digit = 7; }
                    else if (c=='8') { digit = 8; }
                    else if (c=='9') { digit = 9; }
                    else {    
                        // end of digits
                        break;
                    };
                    // add digit to width
                    width = ( width*10 ) + digit;
                }
                
                // decimal?
                if (charAt(format,i) == '.') {
                    // what's the decimal width?
                    decimal = 0;
                    while ( 1==1 ) {
                        // get a char   
                        c = charAt( format, ++i );
                        
                        // translate the digit
                        if      (c=='0') { digit = 0; }
                        else if (c=='1') { digit = 1; }
                        else if (c=='2') { digit = 2; }
                        else if (c=='3') { digit = 3; }
                        else if (c=='4') { digit = 4; }
                        else if (c=='5') { digit = 5; }
                        else if (c=='6') { digit = 6; }
                        else if (c=='7') { digit = 7; }
                        else if (c=='8') { digit = 8; }
                        else if (c=='9') { digit = 9; }
                        else {    
                            // end of digits
                            break;
                        };
                        // add digit to width
                        decimal = ( decimal*10 ) + digit;
                    }
                }
                
                if (argNum+1 > args.length) {
                    Eu.fatal("not enough values to print in printf()" );
                }

                // get an argument
                arg = args.data[argNum++];

                // what type is the argument?                
                if (c=='d') {
                    // decimal number
                    // needs to handle sequences - use first element.
                    sb.append( arg );
                    }    

                if (c=='f') {
                    // floating number
                    // needs to handle sequences, like decimal
                    sb.append( arg );
                    }    

                    
                else if (c=='s') {            
                    if (arg instanceof Seq) {
                        // no width?
                        if (width == 0) {
                            // default
                            width = ((Seq)arg).length;
                        };
    
                        // check size against width
                        for ( int j=0; j < width; j++ ) {
                            // past end of string?    
                            if (j >= ((Seq)arg).length) {
                                // insert space
                                sb.append( ' ' );
                                }
                            else {
                                // insert char
                                sb.append( charAt( (Seq)arg, j ) );
                            }
                        }
                    }
                    else {
                        // insert char
                        if (arg instanceof Integer) {
                            sb.append( (char) ((Integer)arg).doubleValue() );
                            }
                        else {
                            sb.append( (char) ((Double)arg).doubleValue() );
                        }
                        // pad remaining
                        for ( int j=1; j < width; j++ ) {
                                // insert space
                                sb.append( ' ' ); 
                        }
                    }
                }
                
            }
            else {
                // accumulate
                sb.append( c );
            }

        }
        return sb;
    }



    static char charAt( Seq s, int i ) {
        // get the indexed char in the sequence
        Object o;

        // get the Object
        o = s.data[i];

        // convert to char based on class
        if (o instanceof Seq) {
            Eu.fatal( "sequence found inside character string" );
            return ' ';
        }
        else if (o instanceof Integer) {
            return (char)((Integer)o).doubleValue();
            }
        else {
            return (char)((Long)o).doubleValue();
        }
    }


  //## Added support for files
  //## Mic, 000228
    static void print( Object handle, Object e ) {
      // handle sequences
      double val;
      int handle_int;
      handle_int = (int)Eu.toNumber( handle );
      if (handle_int == -1)
	return;
      try {
	if (e instanceof Seq) {
	  // format sequence
	  System.out.println( formatPrint( (Seq)e ) );
	}
	else {   
	  // number
	  val = Eu.toNumber(e);
	  if ( Math.floor(val) == val )
	    // coerce to int (should probably be longer data type)
	    if (handle_int < FIRST_FILE_NUMBER)
	      System.out.println( (int)val );
	    else {
	      /* Print to file */	
	      Integer _temp_integer = new Integer( (int)val );
	      _all_files[handle_int-FIRST_FILE_NUMBER].writeBytes( _temp_integer.toString() );
	    }
	  else                    
	    // display as double
	    if (handle_int < FIRST_FILE_NUMBER)
   	       System.out.println( val );
	    else {
	      /* Print to file */
              Double wrap_val = new Double( val );	
	      _all_files[handle_int-FIRST_FILE_NUMBER].writeBytes( wrap_val.toString() );
	    }
        }
      }
      catch(IOException ioe) {
	Eu.fatal("print(): IO Exception " + ioe);
      }
    }



    // sprintf()
    static Object sprintf( Object format, Object args ) {
        // create a new sequence
        Seq s = new Seq( formatSprintf( format, args ).toString() );
        // set the refcount to zero
        s.ref = 0;
        return (Object) s;
    }
        

  //## Added support for files
  //## Mic, 000302
  static void printf( Object handle, Object format, Object args ) {
    // formatSprintf does the major work
      int handle_int = (int)Eu.toNumber( handle );
      if (handle_int < 0) 
	Eu.fatal( "printf(): Bad handle");
      try {
	if (handle_int < FIRST_FILE_NUMBER) 
	  System.out.print( formatSprintf( format, args ) );
	else
	  _all_files[handle_int-FIRST_FILE_NUMBER].writeBytes(formatSprintf( format, args ).toString());
      }
      catch(IOException ioe) {
	Eu.fatal( "printf(): IO Exception " + ioe );
      }
  }



  // puts()
  //## Added support for files
  //## Mic, 000228
    static void puts( Object handle, Object e ) {
      int handle_int = (int)Eu.toNumber(handle);	
      if (handle_int < 0)
	Eu.fatal("puts(): Bad file handle");

      try {
	if (e instanceof Seq) {
	  // format
	  if (handle_int < FIRST_FILE_NUMBER)
	    System.out.print( Eu.seqToStringBuffer( (Seq)e ) );
	  else
	    _all_files[handle_int - FIRST_FILE_NUMBER].writeBytes(Eu.seqToStringBuffer((Seq)e).toString());						  
	}
	else {       
	  // print
	  if (handle_int < FIRST_FILE_NUMBER)
	    System.out.print( Eu.toNumber(e) );
	  else
	    _all_files[handle_int - FIRST_FILE_NUMBER].writeByte( (int)Eu.toNumber(e) );
	}
      }
      catch(IOException ioe) {
	Eu.fatal("puts(): IO exception "+ ioe);
      }
    }


    // format for puts() statement
    static StringBuffer seqToStringBuffer( Seq s ) {
        // convert to characters for puts
        int c;
        StringBuffer sb = new StringBuffer();

        // set the buffer size to match the Object        
        sb.setLength( s.length );                
        
        // traverse the Object
        for ( int i=0; i < s.length; i++ ) {
        
            // get a character as an int
            c = (int) charAt( s, i );
            
            // cast to char and store in buffer            
            sb.setCharAt( i, (char) c );
        }
               
        return sb;
    }

//////////////////////////////////////////////////////////////////////////////
// Assorted Euphoria routines
//////////////////////////////////////////////////////////////////////////////


    // time()
    static Object time() {
        // cast to a double
        return new Double( ((double)System.currentTimeMillis()/1000) - (double)Eu.startTime );
    }

    // get_key() 
    // !!! this is a hack !!!
    static Object get_key() {
        try { 
            return Eu.create( (int)System.in.read() ); 
        }
        catch (Exception e) {
            Eu.fatal("get_key: fatal exception");
            return null;
        }
        }





//////////////////////////////////////////////////////////////////////////////
// File/device IO
//////////////////////////////////////////////////////////////////////////////

    //## Added open()
    //## Mic, 000227	
    /* Only handles access-modes "r", "w", "rb" and "wb" atm */	
    static Object open(Object a, Object b) {
	/* Make sure both parameters are sequences */
	if ((!(a instanceof Seq)) || (!(b instanceof Seq))) {
		Eu.fatal("open(): Both arguments must be of type sequence");
	}
	
	String a_string = Eu.toString(a);

	try {
		/* Don't do anything if a device-name is encountered */
		if (a_string.equalsIgnoreCase("CON")) {
		} else
		if (a_string.equalsIgnoreCase("AUX")) {
		} else	
		if (a_string.equalsIgnoreCase("COM1")) {
		} else
		if (a_string.equalsIgnoreCase("COM2")) {
		} else
		if (a_string.equalsIgnoreCase("PRN")) {
		} else
		if (a_string.equalsIgnoreCase("NUL")) {
		} else {
			/* Assume a_string is a file on disk */
			if (_file_number < MAX_FILES) {
				String b_string = Eu.toString(b);
				/* Convert the access-mode to something understandable for
				   the RandomAccessFile-object */
				if (b_string.equalsIgnoreCase("r") ||
				    b_string.equalsIgnoreCase("rb")) {
					b_string = "r";
				} else 
				if (b_string.equalsIgnoreCase("w") ||
				    b_string.equalsIgnoreCase("wb")) {
					b_string = "rw";
				}

				RandomAccessFile _new_file =  new RandomAccessFile(a_string, b_string);
				/* Add the File-object to _all_files[] */
				_all_files[_file_number] = _new_file;
				/* Return an index to _all_files[] */
				return Eu.create((_file_number++) + FIRST_FILE_NUMBER);
			}
		}
	}
	catch(IOException e) {
	  /* Return a bad file-number (-1) */
	  return Eu.create( -1 );
	}
	/* Put a return here to make the compiler happy */ 	
	return Eu.create(-1);
    }

  		
    //## Added close(
    //## Mic, 000227	
    static void close(Object a) {
	int a_int = (int)Eu.toNumber(a);

	/* Only try to close if handle != -1 */
	if (a_int != -1) {
	  try {
	    if ((a_int >= FIRST_FILE_NUMBER) && (a_int < FIRST_FILE_NUMBER+MAX_FILES))
		/* Close the file */	
		_all_files[a_int - FIRST_FILE_NUMBER].close();
	  }
	  catch(IOException e) {
	    /* No biggy if close() fails, so just return */	
	    return;
	  }
	}
    }


  //## Added getc()
  //## Mic, 000228
  static Object getc( Object handle ) {
    int handle_int = (int)Eu.toNumber(handle);
    if (handle_int < 0)
      return Eu.create( -1 );

    try {
      if (handle_int < FIRST_FILE_NUMBER)
	/* Dummy code */
	return Eu.create( 0 );
      else
	return Eu.create( (int)(_all_files[handle_int - FIRST_FILE_NUMBER].readByte()) );
    }
    catch(IOException ioe) {
      return Eu.create( -1 );
    }
  }


  //## Added gets()
  //## Mic, 000302
  /* Only handles input from STDIN atm */
  static Object gets( Object handle ) {
    int handle_int = (int)Eu.toNumber(handle);
    if (handle_int < 0)
      return Eu.create( -1 );
    try {
      if (handle_int == 0)
	return new Seq( inData.readLine() + "\r" );
      else
	return "";
    }
    catch(IOException ioe) {
      Eu.fatal( "gets(): IO Exception " + ioe );
    }
    return Eu.create( -1 );   // Return garbage
  }




//////////////////////////////////////////////////////////////////////////////
// Machine functions								
//////////////////////////////////////////////////////////////////////////////


  //## Added machine-function m_seek
  //## Mic, 000228
  /* Includes error-checking-a-plenty wich slows it down a bit */
  static Object m_seek(Seq parms) {
	Object o;
	int handle, pos;
	/* parms should be {fn, pos} */
	o = parms.data[0];
    	if (o instanceof Seq)
      		Eu.fatal("machine_func() (M_SEEK): Expected type integer, found sequence.");
    	handle = (int)Eu.toNumber( o );
    	o = parms.data[1];
    	if (o instanceof Seq)
      		Eu.fatal("machine_func() (M_SEEK): Expected type integer, found sequence.");
    	pos = (int)Eu.toNumber( o );
    	if (handle < FIRST_FILE_NUMBER)
      		Eu.fatal("machine_func() (M_SEEK): Bad file handle");
    	try {
      		RandomAccessFile _temp_file_obj = _all_files[handle - FIRST_FILE_NUMBER];
      		if (pos == -1) {	 	
			_temp_file_obj.seek( _temp_file_obj.length() );
      		} else {
			_temp_file_obj.seek( pos );
      		}
      		return Eu.create( _temp_file_obj.getFilePointer() );     
    	}
    	catch(IOException ioe) {
      		Eu.fatal("machine_func() (M_SEEK): IO Exception " + ioe);
    	}
    	return Eu.create( -1 );	  
  }
	

  //## Added machine-function m_where
  //## Mic, 000228
  static Object m_where(Object parms) {
    int handle;
    	if (parms instanceof Seq) 
      		Eu.fatal("machine_func() (M_WHERE): Expected type integer, found sequence");
      	handle = (int)Eu.toNumber( parms );
      	if (handle < FIRST_FILE_NUMBER)
		Eu.fatal("machine_func() (M_WHERE): Bad file handle");
      	try {
		return Eu.create( _all_files[handle - FIRST_FILE_NUMBER].getFilePointer() );
      	}
      	catch(IOException ioe) {
		Eu.fatal("machine_func() (M_WHERE): IO Exception " + ioe);
      	}
      	return Eu.create( -1 );
    }
  

  //## Added machine_func() 
  //## Mic, 000228
  /* Fails badly if the first parameter is a sequence */	
  static Object machine_func(Object a, Object b) {
	switch((int)Eu.toNumber(a)) {
		case M_SEEK:
			return Eu.m_seek((Seq)b);
			//break;
		case M_WHERE:
			return Eu.m_where(b);
			//break;
		default:
			break;
	}
	return Eu.create( 0 );
  }	


}

