// Euphoria sequence data structure

class Seq extends Object {             

    static final int PADSIZE = 16;

    Object data[];      // array of elements
    int length = 0;     // length of filled elements
    int free = 0;       // number of free elements
    int ref = 0;        // reference count


    public Seq() {
        // create a sequence with default padding
        this.data = new Object[PADSIZE];
        this.free = PADSIZE;
    }
        
    public Seq( int i ) {
        // create a sequence of the requested size
        this.data = new Object[i+PADSIZE];
        this.length = i;
        this.free = PADSIZE;
    }

    public Seq ( String s ) {   
        // create a sequence to hold a string literal
        this.data = new Object[s.length()];
        this.length = s.length();  

        // copy the string        
        for ( int i = 0; i < s.length(); i++ ) {
            this.data[i] = new Integer(  (int)s.charAt(i) );
        }
        
        // since it's a literal, give it a ref
        this.ref = 100;
    };                                                  

    // deref()
    public void deref() {
        // decrease the refcount on a sequence
        
        // decrement refcount
        this.ref -= 1;
    
        // has the refcount fallen to zero?
        if (this.ref < 1) {
        
            // dereference all the sequences in this sequence
            for ( int i = 0; i < this.length; i++ ) {
                // element is a sequence?
                if (this.data[i] instanceof Seq) {
                    // deref
                    ((Seq)this.data[i]).deref();
                    // force? 
                    //((Seq)this.data[i]) = null;
                    
                }
            }
        }
    }


    public Seq makeUnique() {
        // if the sequence is not unique, return a
        // shallow copy of the sequence
        Seq sCopy;
        Object element;

        // unique?
        if (this.ref == 0 ) {
            // don't clone it
            return this;
            
        } else {

            // create new sequence
            sCopy = new Seq(this.length);
            
            // set attributes
            sCopy.ref = 0;
            sCopy.length = this.length;
            sCopy.free = PADSIZE;
            
            // copy the elements
            for ( int i = 0; i < this.length; i++ ) {
                // get an element
                element = this.data[i];
                
                // place in sequence            
                sCopy.data[i] = element;
                
                // refcount?
                if (element instanceof Seq) {
                    // increase refcount
                    ((Seq)element).ref += 1;
                }
            }
                
            return sCopy;
        }
    }

    public void addElement( Object o ) {
        // add an element onto the sequence

        // needs more room?
        if (this.free == 0) {

            // amount needed
            int len = this.length+PADSIZE;
        
            // create a larger sequence
            Object tmp[] = new Object[ len ];
            
			// copy elements over
			System.arraycopy(data,0,tmp,0,this.length);

            // replace sequence
            this.data = tmp;               
            
            // update free size
            this.free = PADSIZE;
        }

        // add element
        this.data[ this.length ] = o;

        // increase the length
        this.length++;                     

        // decrement free
        this.free--;
    }
    
}

