-- Euphoria to Java translation
-- ej.e

--include ox_stub.e

-- support routines
include get.e
include wildcard.e

sequence out
out = repeat( {}, 4 )
integer emitTo
integer lastReturnAt
constant
    ToRoutine   = 1,
    ToModule    = 2,
    ToTmp       = 3
    
    emitTo = ToModule
    lastReturnAt = 0

global constant         
    NoOp            = 0,
    DefProc         = 1,
    DefFunc         = 2,
    DefType         = 3,
    CallProc        = 4,
    CallFunc        = 5,
    Return          = 6,
    If              = 7,
    Elsif           = 8,
    Else            = 9,
    End             = 10,
    For             = 11,
    While           = 12,
    GetIndex        = 13,
    SetIndex        = 14,
    Slice           = 15,
    Literal         = 16,
    CreateVar       = 17,
    CreateConst     = 18,
    CreateSeq       = 19,
    Assign          = 20,
    BinaryOp        = 21,
    NumLiteral      = 22,
    StrLiteral      = 23,
    Variable        = 24,
    Negate          = 25,
    Compare         = 26,
    Not             = 27,
    And             = 28,
    Or              = 29,
    Xor             = 30,
    ModuleEnd       = 31,
    Exit            = 32,
    Include         = 33,
    GlobalFlag      = 34

sequence opCode
    opCode = repeat( -1, 50 )

sequence euPath


global constant
    Function        = 1,
    Procedure       = 2,
    Type            = 3
    
global constant        
    SET             = 0,
    ADD             = 1,
    SUB             = 2,
    MUL             = 3,
    DIV             = 4,
    CONCAT          = 5,
    FLOOR_DIV       = 6,

    EQ              = 1,
    NE              = 2,
    LT              = 3,
    LE              = 4,
    GT              = 5,
    GE              = 6    

constant
    True            = not 0,
    False           = 0,
    SeqPadding      = 10            -- extra padding on sequences


-- order matches search order
global constant
    Global  = 1,
    Module  = 2,
    Routine = 3,
    SearchOrder = { Routine, Global, Module }

-- information in routines
global constant
    Name            = 1,        -- euphoria name of routine
    JavaName        = 2,        -- java name of routine
    Scope           = 3,        -- scope id flag
    Returns         = 4,        -- return type
    Code            = 5,        -- code to emit
    Args            = 6,        -- arg list
    Touches         = 7,        -- list of routines it references
    IsTouched       = 8,        -- if true, is called
    IsRoutineId     = 9,        -- if true, called by routine_id
    SizeOfRoutine   = 9

sequence
    modules,            -- list of module names
    moduleCode,         -- commands in module               
    moduleVars,         -- holds names of all module variables
    pendCode,           -- function/procedure code body
    mainCode,           -- code that goes in the main module    
    local,              -- list of locals
    derefList,          -- arglist, for dereferencing
    anon,               -- anonymous variables
    stack,              -- temp stack
    lits,               -- literals
    types,              -- names of types
    typedefs,           -- what types translate to
    token,              -- name of tokens
    token_args,         -- args taken by token
    token_type,         -- datatype of token
    var,                -- names of vars          
    prefix,             -- prefix to add to module
    included,           -- list of include files
    class,              -- class name
    jimports,           -- java imports
    touches,            -- list of items current routine calls
    routine,            -- information on converted routines    
    routine_ids         -- routines called by routine_id

    modules     = {}
    moduleCode  = {}
    pendCode    = {}
    mainCode    = {}
    local       = {}
    anon        = {}
    stack       = {}
    lits        = {}    
    types       = {{ "object", "sequence", "atom", "integer" }, {}, {}}
    typedefs    = {{ "object", "sequence", "atom", "integer" }, {}, {}}
    var         = repeat( {}, 3 )
    token       = repeat( {}, 3 )
    token_args  = repeat( {}, 3 )
    token_type  = repeat( {}, 3 )
    prefix      = { "_0", "_1", "_2" }
    moduleVars  = repeat( {}, 3 )
    included    = {}
    routine_ids = {}
    jimports    = {}
    touches     = {}
    routine     = {}
integer
    prefixCount,
    usesRoutineId
    
    prefixCount     = 1     --  for uniquely identifying modules
    usesRoutineId   = 0     -- if false, never calls routine_id

global integer 
    globalFlag,
    theScope,
    isFunction

    globalFlag = 0
    theScope = Module
    isFunction = 0

sequence builtin
builtin = {
    -- debugging
    "debug_ref",    Procedure,  1,  "System.out.println( \"debug_ref:\" + ((Seq)%s).ref )",
    
    -- euphoria builtins
    "abort",        Procedure,  1,  "Eu.abort( %s )",
    "and",          Function,   2,  "Eu.binaryOp( Eu.AND, %s, %s )",
    "and_bits",     Function,   2,  "Eu.binaryOp( Eu.AND_BITS, %s, %s )",
    "append",       Function,   2,  "Eu.append( %s, %s )",
    "arctan",       Function,   1,  "Eu.unaryOp( Eu.ARCTAN, %s )",
    "atom",         Function,   1,  "Eu.isAtom( %s )",
    "call_func",    Function,   2,  "class.call_func( %s, %s )",
    "call_proc",    Procedure,  2,  "class.call_proc( %s, %s )",
    "clear_screen", Procedure,  0,  "Eu.dummyProc()",                   -- dummy
    --## Added close()
    --## Mic, 000227
    "close",        Procedure,  1,  "Eu.close( %s )",
    --## Added command_line()
    --## Mic, 000303
    "command_line", Function,   0,  "new Seq()",	-- dummy
    "compare",      Function,   2,  "Eu.compare( %s, %s )",
--    "concat",       2,  "Eu.concat( %s, %s )",
    "cos",          Function,   1,  "Eu.unaryOp( Eu.COS, %s )",
    "equal",        Function,   2,  "Eu.equal( %s, %s )",
    "find",         Function,   2,  "Eu.find( %s, %s )",
    "floor",        Function,   1,  "Eu.unaryOp( Eu.FLOOR, %s )",
    "get_key",      Function,   0,  "Eu.get_key()",                 -- only stdin
    --## Added getc()
    --## Mic, 000228	
    "getc",         Function,   1,  "Eu.getc( %s )",		-- only for files 
    --## Added getenv()
    --## Mic, 000303	
    "getenv",       Function,   1,  "Eu.create( -1 )",	-- dummy 
    --## Added gets()
    --## Mic, 000228	
    "gets",         Function,   1,  "Eu.gets( %s )",		-- only stdin 
    "integer",      Function,   1,  "Eu.isInteger( %s )",
    "length",       Function,   1,  "Eu.length( %s )",
    "log",          Function,   1,  "Eu.unaryOp( Eu.LOG, %s )",
    --## Added machine_func()
    --## Mic, 000228
    "machine_func", Function,   2,  "Eu.machine_func( %s, %s )", 
    "machine_proc", Procedure,  2,  "Eu.dummyProc()",               -- dummy
    --## Added match()  (not yet implemented)
    --## Mic, 000303
    "match",        Function,   2,  "Eu.match( %s, %s )", 
    "not",          Function,   1,  "Eu.unaryOp( Eu.NOT, %s )",
    "not_bits",     Function,   1,  "Eu.unaryOp( Eu.NOT_BITS, %s )",
    "object",       Function,   1,  "Eu.create( 1 )",
    --## Added open()
    --## Mic, 000227
    "open",	    Function,   2,  "Eu.open( %s, %s )",
    "or",           Function,   2,  "Eu.binaryOp( Eu.OR, %s, %s )",
    "or_bits",      Function,   2,  "Eu.binaryOp( Eu.OR_BITS, %s, %s )",
    --## Added platform()
    --## Mic, 000303
    "platform",     Function,   0,  "Eu.create( 0 )",			-- dummy	
    "poke",         Procedure,  2,  "Eu.dummyProc()",                   -- dummy
    "position",     Function,   2,  "Eu.dummyProc()",                   -- dummy
    "power",        Function,   2,  "Eu.binaryOp( Eu.POWER, %s, %s )",
    "prepend",      Function,   2,  "Eu.prepend( %s, %s )",
    "print",        Procedure,  2,  "Eu.print( %s, %s )",
    "printf",       Procedure,  3,  "Eu.printf( %s, %s, %s )",
    "puts",         Procedure,  2,  "Eu.puts( %s, %s )",
    "rand",         Function,   1,  "Eu.unaryOp( Eu.RAND, %s )",
    "remainder",    Function,   2,  "Eu.binaryOp( Eu.REMAINDER, %s, %s )",
    "repeat",       Function,   2,  "Eu.repeat( %s, %s )",
    "routine_id",   Function,   1,  "class.routine_id( %s, %s )",
    "sequence",     Function,   1,  "Eu.isSequence( %s )",
    "sin",          Function,   1,  "Eu.unaryOp( Eu.SIN, %s )",
    "sprintf",      Function,   2,  "Eu.sprintf( %s, %s )",
    "sqrt",         Function,   1,  "Eu.unaryOp( Eu.SQRT, %s )",
    --## Added system()
    --## Mic, 000303
    "system",       Procedure,  2,  "Eu.dummyProc()",	-- dummy
    "tan",          Function,   1,  "Eu.unaryOp( Eu.TAN, %s )",
    "time",         Function,   0,  "Eu.time()",
    "xor",          Function,   2,  "Eu.binaryOp( Eu.XOR, %s, %s )",
    "xor_bits",     Function,   2,  "Eu.binaryOp( Eu.XOR_BITS, %s, %s )" }

global procedure setClassName( sequence fileName )
    -- convert file name to class name
    integer at      
    
    -- convert filename to class name
    class = fileName
    -- class = lower( fileName )
    -- class[1] -= 'a' - 'A'
    at = find( '.', class )
    if at then
        class = class[1..at-1]
    end if
    
end procedure

function removeItem( sequence s, integer i )
    return s[1..i-1] & s[i+1..length(s)]
end function

function isNumeric( sequence var )
    -- returns true if variable is numeric type
    integer scope, at

    -- don't process if it's a local
    if var[1] = '_' then
        return False
    end if

    -- look through scopes    
    for i = 1 to length( SearchOrder ) do
        scope = SearchOrder[i]
        at = find( var, token[scope] )
        if at then
            if equal( token_type[scope][at], "integer" )
            or equal( token_type[scope][at], "atom" ) then
                return True
            else
                return False
            end if
        end if
    end for        

    oxParseError( "%s has not been declared", {var} )
    return "_error_"
end function


function javaName( sequence word )
    -- returns internal name of token
    integer scope
    for i = 1 to length( SearchOrder ) do
        scope = SearchOrder[i]
        if find( word, token[scope] ) then
            return prefix[scope] & word
        end if
    end for        

    oxParseError( "%s has not been declared", {word} )
    return "_error_"
end function


function defined( sequence word )
    -- returns true if word is defined
    integer scope

    for i = 1 to length( SearchOrder ) do
        scope = SearchOrder[i]
        if find( word, token[scope] ) then
            return 1
        end if
    end for

    -- builtin?
    for i = 1 to length( builtin ) by 4 do
        if equal( word, builtin[i] ) then
            return 1
        end if
    end for


    return 0

end function


function defineRoutine( integer scope, sequence word, integer argCount )
    -- add the routine token to the list
    integer id
    sequence info

    -- builtin?
    for i = 1 to length( builtin ) by 4 do
        if equal( word, builtin[i] ) then                        
            -- warning
            printf( 1, "Warning: built-in %s redefined", {word} )
            -- remove
            builtin[i] = ""
            exit
        end if
    end for
    
    -- not already on list?
    if  not find( word, token[scope] )
    and not find( word, token[Global] ) then
    
        -- add it
        token[scope] &= {word}
        token_args[scope] &= argCount
        token_type[scope] &= {"_"}

        -- figure out the id
        if scope = Global then
            -- global
            id = 0
        else                    
            -- module
            id = prefixCount
        end if

        -- create record
        info = repeat( 0, SizeOfRoutine )
        info[Name]      = word
        info[JavaName]  = prefix[scope] & word
        info[Scope]     = prefix[scope]
        
        -- add to routines
        routine &= { info }
        
        -- return index
        return length( routine )
        

    else
        -- error
        oxParseError( "attempt to redefine %s", {word} )
    end if       
    
end function

procedure define( integer scope, sequence word )
    -- adds the token to the defined token list
    
    -- not already on list?
    if not find( word, token[scope] ) then
        -- add it
        token[scope] &= {word}
        token_args[scope] &= -1
        token_type[scope] &= {"_"}
    end if       
    
end procedure

procedure defineVar( integer scope, sequence word, sequence typeof )
    -- adds the token to the defined token list
    
    -- not already on list?
    if not find( word, token[scope] ) then
        -- add it
        token[scope] &= {word}
        token_args[scope] &= -1
        token_type[scope] &= {typeof}
        var[scope] &= {word}
        
        -- module level?
        if scope = Module
        and not globalFlag then
            -- add to module variable list
            moduleVars[prefixCount] &= { prefix[scope] & word }
        end if
        
    end if       
    
end procedure

procedure forget( integer scope, sequence word )
    -- forget a token
    integer at
    
    at = find( word, token[scope] )
    if at = 0 then
        oxParseError( "%s has not been declared", {word} )
        return    
    end if   
    
    -- remove
    token[scope]      = removeItem( token[scope], at )
    token_args[scope] = removeItem( token_args[scope], at )
    token_type[scope] = removeItem( token_type[scope], at )

    -- variable?
    at = find( word, var[scope] )
    if at then
        var[scope] = removeItem( var[scope], at )
    end if
    
end procedure

procedure forgetAll( integer scope )
    -- forget all tokens within scope
    sequence s

    -- tokens
    token[scope] = {}
    token_args[scope] = {}
    token_type[scope] = {}
    var[scope] = {}
    types[scope] = {}
    typedefs[scope] = {}

end procedure


procedure checkArgCount( sequence word, integer count )
    -- returns internal name of token 
    integer at, takes, scope
    
    for i = 1 to length( SearchOrder ) do
        scope = SearchOrder[i]
        at = find( word, token[scope] )
        if at then   
            takes = token_args[scope][at]
            if count != takes then
                oxParseError( "%s takes %d arguments", {word, takes} )
            end if
            return
        end if
    end for        
    
    -- builtin?
    for i = 1 to length( builtin ) by 4 do
        if equal( word, builtin[i] ) then
            takes = builtin[i+2]
            if count != takes then
                oxParseError( "%s takes %d arguments", {word, takes} )
            end if
            return
        end if
    end for
    
    oxParseError( "%s is not defined\n", {word} )
end procedure

function tab( integer i )  
    -- return requested number of spaces
    return repeat( ' ', i )
end function

global function val( object o )
    -- return numeric value of o
    o = value( o )
    return o[2]
end function

function evalWord( sequence word )
    integer at, scope
    sequence val
    
    -- evaluate a word, and sets the token and return values
    if find( word, oxLiteral )
    or    find( word, oxKeyword ) then
        -- return as a literal
        return { word, word }

    elsif word[1] = '"' then
        -- string                  
        return { "STRING", word }
        
    elsif word[1] = '\'' then
        -- character literal        
        return { "NUMBER", word }
        

    elsif word[1] >= '0' and word[1] <= '9' then
        -- number
        return { "NUMBER", word }

    elsif word[1] = '#' then
        -- convert
        val = value( word )
        return { "NUMBER", sprintf("%d", val[2]) }


    else
        -- type?       
        for i = 1 to length( SearchOrder ) do
            -- get scope
            scope = SearchOrder[i]
            -- look in scope
            if find( word, types[scope] ) then
                return { "TYPE", word }
            end if
        end for
        
        return { "IDENTIFIER", word }
        
    end if
    
end function

oxSetEvalWord( routine_id("evalWord" ) )


function getValue( sequence s )
    -- convert s to a numeric value
    s = value( s )
    return s[2]
end function


global function createVar( integer isGlobal, sequence typeOf, sequence name )
    -- declare a variable in the current namespace
    -- returns the name of the variable in C
    integer at, scope

    -- global?
    if isGlobal then
        -- use global scope
        scope = Global  
        
    else
        -- use current scope (module or routine)
        scope = theScope
        
    end if

    if not find( name, token[scope] ) then
        -- define it
        defineVar( scope, name, typeOf )
    end if

    -- name of variable
    return javaName( name )
    
end function


procedure emit( integer tab, sequence text, sequence args )
    -- emit code
    integer handle 
    text &= "\n"
    -- printf( 1, text, args )
    
    if tab then
        out[emitTo] &= repeat( ' ', tab )
    end if
    
    out[emitTo] &= sprintf( text, args )
end procedure


procedure comment( sequence text, sequence args )
    emit( 0, "", {} )
    emit( 8, "// " & text, args )
end procedure


procedure ref( sequence name )
    -- increase refcount on object
    
    -- old version
    -- emit( 8, "Eu.ref( %s );", {name} )

    -- can skip if it's and int or atom
    if isNumeric( name ) then
        -- nothing to do
    else
        -- inline the code
        emit( 8, "if (%s instanceof Seq) {", {name} )
        emit( 8, "    // increase the refcount", {} )
        emit( 8, "    ((Seq)%s).ref += 1;", {name} )
        emit( 8, "}", {} )
    end if
    
end procedure

procedure deref( sequence name )
    -- decrease the reference count on an object
    
    -- old version
    -- emit( 8, "Eu.deref( %s );", {name} )
    
    -- can skip if it's and int or atom
    if isNumeric( name ) then
        -- nothing to do
    else
        -- inline the code
        emit( 8, "if (%s instanceof Seq) {", {name} )
        emit( 8, "    // decrease refcount", {} )
        emit( 8, "    ((Seq)%s).deref();", {name} )
        emit( 8, "}", {} )
    end if
    
end procedure
    

sequence freeAnonList, usedAnonList, literalList, literalName
            
    freeAnonList = {}
    usedAnonList = {}
    literalList = {}
    literalName = {}

function createAnon()
    -- create a temp variable
    -- first, look for a free one

    -- no free anon vars?
    if length( freeAnonList ) = 0 then
        -- add a new one to the pool
        freeAnonList = { sprintf( "_tmp%d", length( usedAnonList ) ) }
    end if
    
    -- pick an first anon from the free list
    usedAnonList = prepend( usedAnonList, freeAnonList[1] )
    freeAnonList = freeAnonList[2..length(freeAnonList)]
    
    return usedAnonList[1]
    
end function

function createLiteral( sequence val )

    -- create a literal, if it doesn't already exist
    integer at
    sequence name


    -- exists?    
    at = find( val, literalList )
    if at = 0 then
        -- add to literals    
        literalList = append( literalList, val )
        at = length( literalList )
        
        -- create name
        if val[1] = '\"' then
            name = sprintf( "_string%d", {at} )
        else
            name = sprintf( "_literal%d", {at} )
        end if
        literalName &= {name}

    end if

    return literalName[at]
    
end function


procedure freeAnon( integer derefFlag, sequence list )
    -- if the item is in the anonymous variable list,
    -- free it
    integer at, count
    sequence name

    count = 0
    for i = 1 to length( list ) do
        name = list[i]
        
        at = find( name, usedAnonList )
        if at then

            -- copy to free list
            freeAnonList = append( freeAnonList, usedAnonList[at] )
            
            -- remove from used list
            usedAnonList = removeItem( usedAnonList, at )

            -- increase count        
            count += 1
             
            if count = 1 then
                comment( "garbage collection", {} )
            end if

            -- deref?
            if derefFlag then   
                deref( name )
            end if
            
            -- free
            emit( 8, "%s = null;", {name} )
            
        end if
        
    end for

    -- any variables freed?
    if count then
        -- suggest garbage collection
--        emit( 8, "System.gc();", {} )
    end if
    
    
end procedure


-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
        
global function eval( sequence code )
    integer op
    op = code[1]

    if op = NoOp then
        -- { NoOp }
        return True
    else
        return call_func( opCode[ code[1] ], {code} )
    end if
    
end function
    

function opCreateVar( sequence code )
    -- { CreateVar, type, var1, var2 ... }
    sequence name    
    
    -- create the variables
    for i = 3 to length( code ) do
        name = createVar( globalFlag, code[2], code[i] )
    end for      
    
    -- clear the global flag
    globalFlag = 0
    
    return True
    
end function
opCode[CreateVar] = routine_id("opCreateVar")

function opCreateConst( sequence code )
    -- { CreateConst, name, expr, [name, expr...] }
    sequence tmp, name

    for i = 2 to length( code ) by 2 do
        name = code[i]
        comment( "assign to constant %s", {name} )        
        emit( 8, "%s = %s;", {javaName(name), eval( code[i+1] )} )
        ref( javaName(name) )
    end for
    
    return True    
    
end function
opCode[CreateConst] = routine_id("opCreateConst")


function opCreateSeq( sequence code )
    -- { CreateSeq, expr1, expr2 ... }
    integer used
    sequence tmp, expr, tmp2

    tmp = createAnon()
    used = 0

    if length( code ) = 1 then
        comment( "create an empty sequence", {} )
        emit( 8, "%s = (Object) new Seq();", {tmp} )
    else
        comment( "create a new sequence", {} )
        emit( 8, "%s = (Object) new Seq(%d);", {tmp,length(code)-1} )
    
        for i = 2 to length( code ) do

            -- eval the expression
            expr = eval( code[i] )
            
            -- not a numeric literal?
            if expr[2] != 'l' then
                -- need tmp?
                if not used then
                    tmp2 = createAnon()
                end if
                
                -- place in temp and ref
                comment( "attach element %d", {i-1} )
                emit( 8, "%s = %s;", {tmp2,expr} )
                emit( 8, "((Seq)%s).data[%d] = %s;", {tmp,i-2,tmp2} )
                ref( tmp2 )
                
                -- flag tmp2 was used
                used = 1
            else
                -- assign directly, no ref
                emit( 8, "((Seq)%s).data[%d] = %s;", {tmp,i-2,expr} )
            end if
            
        end for         

        -- free up tmps        
        if used then
            freeAnon( 0, {expr, tmp2} )
        else
            freeAnon( 0, {expr} )
        end if
        
    end if

    return tmp

end function
opCode[CreateSeq] = routine_id("opCreateSeq")

    
function opAssign( sequence code )
    -- { Assign, var, Expr }
    sequence var, expr, tmp
    
    -- get the var name
    var = javaName( code[2] )

    -- numeric?
    if isNumeric( code[2] ) then
        expr = eval( code[3] )
    
        -- simple assignment
        comment( "assign number to %s", {code[2]} )
        emit( 8, "%s = %s;", {var, expr} )
        freeAnon( 0, {expr} )
    else
        -- need to dereference prior value
    
        -- get an anon var
        tmp = createAnon()
        expr = eval( code[3] )
                
        -- replace with evaluated expression
        comment( "assign to %s", {code[2]} )
        emit( 8, "%s = %s;", {tmp, var} )
        emit( 8, "%s = %s;", {var, expr} )
        ref( var )
        
        -- dereference original value
        deref( tmp )
        
        -- free local
        freeAnon( 0, {expr, tmp} )
        
        
    end if
    
    return True

end function
opCode[Assign] = routine_id("opAssign")


function opSetIndex( sequence code )
    -- { SetIndex, op, var, indexes, expr }        
    integer j, op, isSlice
    sequence prior, src, indexes, indexOffset, indexData, expr, index,
                name, indexedItem

    -- both setItem and setSlice take care of setting the refcounts
    -- internally, so it doesn't have to be handled in the code.

    indexes = code[4]
    indexOffset = {}
    indexData = {javaName( code[3] )}
    for i = 1 to length( indexes )-1 do

        comment( "indexed store into %s", {code[3]} )

        -- temp variable to hold offset
        indexOffset &= {createAnon()}

        -- evaluate the index
        comment( "evaluate index #%d", {i} )
        emit( 8, "%s = %s;", {indexOffset[i], eval(indexes[i])} )
        
        -- get the data
        comment( "indexed fetch #%d", {i} )

        -- what's the source?
        indexData &= { createAnon() }

        -- fetch
        emit( 8, "%s = Eu.getItem( %s, %s );", 
            {indexData[i+1],indexData[i],indexOffset[i]} )
        
    end for

    -- evaluate the last index
    j = length( indexes )

    -- get the op
    op = code[2]

    -- slice?
    isSlice = (indexes[j][1] = Slice)
    
    -- what's the operation?
    if op = CONCAT then name = "CONCAT"
    elsif op = SET then name = "SET"
    elsif op = ADD then name = "ADD"
    elsif op = SUB then name = "SUB"
    elsif op = MUL then name = "MUL"
    elsif op = DIV then name = "DIV"
    end if


    -- is the last index a slice?
    if isSlice then
        if op = CONCAT then      
            -- illegal operation?
            -- this message is different than RDS'
            oxParseError( "can't concatonate to a slice", {} )
            return True
        end if

        -- temp variable to hold offset
        indexOffset &= {createAnon()}
        indexOffset &= {createAnon()}
        comment( "slice low index", {} )
        emit( 8, "%s = %s;", {indexOffset[j],eval(indexes[j][2])})
        
        comment( "slice high index", {} )
        emit( 8, "%s = %s;", {indexOffset[j+1],eval(indexes[j][3])})

        -- perform the operation
        comment( "assign the slice", {} )
        emit( 8, "%s = Eu.setSlice( Eu.%s, %s, %s, %s, %s );", 
            {indexData[j],name,indexData[j],indexOffset[j],indexOffset[j+1],
                eval(code[5])} )
    
    else

        -- temp variable to hold offset
        indexOffset &= {createAnon()}
        comment( "evaluate the index", {} )
        emit( 8, "%s = %s;", {indexOffset[j],eval(indexes[j])})
    
        -- perform the operation                  
        emit( 8, "%s = Eu.setItem( Eu.%s, %s, %s, %s );", 
            {indexData[j],name,indexData[j],indexOffset[j],eval(code[5])} )
    
    end if
        
    -- nested?
    if j > 1 then
        -- unroll
        for i = j-1 to 1 by -1 do
            -- assign                    
            comment( "set the indexed item", {} )
            emit( 8, "%s = Eu.setItem( Eu.SET, %s, %s, %s );",
                {indexData[i], indexData[i], indexOffset[i], indexData[i+1] } )
        end for
    end if

    -- free locals
    freeAnon( 0, indexOffset )
    freeAnon( 1, indexData )
    
    return True

end function
opCode[SetIndex] = routine_id("opSetIndex")



function opBinaryOp( sequence code )
    -- { BinaryOp, code, left, right }
    integer op
    sequence name, long, left, right, tmp
    op = code[2]

    -- determine the op to use
    if    op = ADD then 
        name = "ADD"             
        long = "add"
    elsif op = SUB then 
        name   = "SUB"   
        long   = "subtract"
    elsif op = MUL then 
        name   = "MUL"
        long = "multiply"
    elsif op = DIV then 
        name   = "DIV"           
        long   = "divide"
    elsif op = FLOOR_DIV then 
        name   = "FLOOR_DIV"           
        long   = "integer divide"
    end if

    -- evaluate both sides
    left = eval( code[3] )
    right = eval( code[4] )

    -- holder for result
    tmp = createAnon()

    if op = CONCAT then                                 
        comment( "concat", {} )
        emit( 8, "%s = Eu.concat( %s, %s );", {tmp,left,right} )
    else
        -- perform the operation
        comment( "%s", {long} )                            
        emit( 8, "%s = Eu.binaryOp( Eu.%s, %s, %s );", {tmp,name,left,right} )
    end if
            
    -- free anons
    freeAnon( 1, {left, right} )
    
    return tmp


end function
opCode[BinaryOp] = routine_id("opBinaryOp")


procedure parseToEnd( integer endTag, sequence endTagText )
    
    -- parse until "end" encountered
    object result
    
    while 1 do

        -- end of file?
        if oxTokenIs("<EOF>") then
            oxParseError( "Expected 'end %s', not <eof>", {endTagText} )
        else
            -- parse a command
            result = oxParse( "Command" )
            if oxError then
                return
            end if    
            
            if result[1] = End then
                if result[2] = endTag then
                    -- leave loop
                    exit     
                else
                    oxParseError( "Expected '%s', not %s", {endTagText,oxToken} )
                end if
            else                       
                -- evaluate code
                result = eval( result )
            end if
        end if
    end while 
    
end procedure

function opFor( sequence code )
    -- { For, varName, start, end, step }         
    sequence loopVar, loopStart, loopEnd, loopStep, javaVar
    object result

    -- get loop variable
    loopVar = code[2]
    javaVar = createVar( 0, "atom", loopVar )

    -- get the loop start
    comment( "loop start", {} )
    loopStart = createAnon()
    emit( 8, "%s = %s;", {loopStart,eval( code[3] )} )

    -- get the loop end
    comment( "loop end", {} )
    loopEnd = createAnon()
    emit( 8, "%s = %s;", {loopEnd,eval( code[4] )} )
    
    -- get the step value
    comment( "loop step", {} )
    loopStep = createAnon()
    emit( 8, "%s = %s;", {loopStep,eval( code[5] )} )

    -- the for loop
    comment( "for %s ... loop", {loopVar} )
    emit( 8, "for(", {} )
    emit( 8, "     // loop range", {} )
    emit( 8, "     double _index_%s = Eu.toNumber( %s ),", {loopVar,loopStart} )
    emit( 8, "            _max_%s   = Eu.toNumber( %s ),", {loopVar,loopEnd} )
    emit( 8, "            _step_%s  = Eu.toNumber( %s );", {loopVar,loopStep} )
    emit( 0, "",{} )
    emit( 8, "     // check for end of loop", {} )
    emit( 8, "     Eu.forLoop( _index_%s, _step_%s, _max_%s );", {loopVar,loopVar,loopVar} )
    emit( 0, "", {} )
    emit( 8, "     // increment loop", {} )
    emit( 8, "     _index_%s += _step_%s ) {", {loopVar,loopVar} )

    comment( "set loop index %s", {loopVar} )
    emit( 8, "%s = Eu.create( _index_%s );", {javaVar,loopVar} )

    -- parse until "end" encountered
    parseToEnd( For, "for" )
    if oxError then
        return 0
    end if
    
    -- bottom of loop
    comment( "end of for %s ... loop", {loopVar} )
    emit( 8, "}", {} )
    
    -- free temps
    freeAnon( 0, {loopStart, loopEnd, loopStep} )

    return True    
    
end function
opCode[For] = routine_id("opFor")

function opCompare( sequence code )
    -- { Compare, op, expr1, expr2 }
    sequence op, left, right, tmp

    -- evaluate
    comment( "evaluate test", {} )
    left = eval( code[3] )
    right = eval( code[4] )

    -- holds result
    tmp = createAnon()

    -- what operation?
    if code[2] = EQ then           op = "EQ"
    elsif code[2] = NE then        op = "NE"
    elsif code[2] = GT then        op = "GT"
    elsif code[2] = GE then        op = "GE"
    elsif code[2] = LT then        op = "LT"
    elsif code[2] = LE then        op = "LE"
    end if

    -- test
    emit( 8, "%s = Eu.binaryOp(Eu.%s, %s, %s);", {tmp,op,left,right} )

    
    -- free items
    freeAnon( 1, {left, right} )

    return tmp

end function
opCode[Compare] = routine_id("opCompare")


function logical( sequence code )
    -- { AND/OR/XOR, expr1, expr2 }
    sequence op, left, right, tmp

    if    code[1] = Or  then op = "OR" 
    elsif code[1] = Xor then op = "XOR"
    elsif code[1] = And then op = "AND"
    end if

    -- evaluate
    comment( "evaluate left and right sides", {} )
    left = eval( code[2] )
    right = eval( code[3] )

    -- holds result
    tmp = createAnon()
    
    -- code
    emit( 8, "%s = Eu.binaryOp(Eu.%s, %s, %s);", {tmp,op,left,right} )
    
    -- free items
    freeAnon( 1, {left, right} )

    return tmp

end function
opCode[Or] = routine_id("logical")
opCode[Xor] = routine_id("logical")
opCode[And] = routine_id("logical")


procedure evalBoolean( sequence code )
    -- evaluate boolean expression, setting _test
    -- converts 'or' and 'and' to short-circuit tests
    sequence tmp

    if code[1] = Or
    or code[1] = And then
        -- { Or, expr, expr }
        -- { And, expr, expr }

        -- evaluate left hand side
        evalBoolean( code[2] )

        -- which conditional?
        if code[1] = Or then
            comment( "short-circuit or", {} )
            emit( 8, "if (!_test) {", {} )
        else
            comment( "short-circuit and", {} )
            emit( 8, "if (_test) {", {} )
        end if
        
        -- evaluate right hand side
        evalBoolean( code[3] )
        emit( 8, "}", {} )
    
    else    
        tmp = eval(code)
        emit( 8, "_test = (Eu.toNumber( %s ) != 0);", {tmp} )
        freeAnon( 0, {tmp} )
    end if
    
end procedure

function opIf( sequence code )
    -- { If, expr }
    integer branch
    object result

    -- no branch yet
    branch = 0

    -- evaluate the if portion
    comment( "if", {} ) 
    evalBoolean( code[2] )
                  
    comment( "branch?", {} )
    emit( 8, "if (_test) {", {})

    -- parse until end if
    while 1 do

        -- end of file?
        if oxTokenIs("<EOF>") then
            oxParseError( "Expected 'end if', not <eof>", {} )
        else
            -- parse a command
            result = oxParse( "Command" )
            if oxError then
                return 0
            end if

            if result[1] = Elsif then      

                -- prior statment not return?
                if lastReturnAt = 0
                or lastReturnAt != length( out[ToTmp] ) then
                    -- clear the flag          
                    comment( "prevent elsif/else branches from executing", {} )
                    emit( 8, "_test = true;", {} )
                end if
                
                -- close block
                emit( 8, "}", {} ) 

                -- close elsif?
                if branch then
                    -- close extra brace
                    emit( 8, "}", {} ) 
                end if
                
                -- flag elsif encountered, they nest two deep
                branch = 1

                -- eval
                comment( "eval elsif", {} )
                emit( 8,  "if (!_test) {", {} )

                comment( "evaluate condition", {} )
                evalBoolean( result[2] )
                
                comment( "branch?", {} )
                emit( 8, "if (_test) {", {})
                
            elsif result[1] = Else then

                -- prior statment not return?
                if lastReturnAt = 0
                or lastReturnAt != length( out[ToTmp] ) then
                    -- clear the flag          
                    comment( "prevent else branch from executing", {} )
                    emit( 8, "_test = true;", {} )
                end if
                
                -- close block
                emit( 8, "}", {} )
                
                -- close elsif?
                if branch then
                    -- close extra brace
                    emit( 8, "}", {} )
                    
                    -- kill flag
                    branch = 0
                end if
            
                comment( "else", {} )
                emit( 8,  "if (!_test) {", {} )
                
            elsif result[1] = End then
                if result[2] = If then                
                    -- leave loop
                    exit     
                else
                    oxParseError( "Expected 'if', not %s", {oxToken} )
                end if
            else                       
                -- evaluate code
                result = eval( result )
            end if
        end if
    end while 

    -- close elsif?
    if branch then
        -- close extra brace
        emit( 8, "}", {} )
    end if

    -- end if
    comment( "end if", {} )
    emit( 8, "}", {})

    return True
    
end function
opCode[If] = routine_id("opIf")


function opWhile( sequence code )
    -- { While, expr }

    -- begin while
    comment( "while", {} )
    emit( 8, "while( true ) {", {} )
    
    -- test
    emit( 8,  "if ( Eu.toNumber( %s ) == 0 ) {", { eval( code[2] ) } )
    emit( 12, "break;", {} )
    emit( 8,  "}", {} )
    
    -- parse
    parseToEnd( While, "while" )
    if oxError then
        return 0
    end if

    -- end while
    comment( "end while", {} )
    emit( 8, "}", {})

    return True
    
end function
opCode[While] = routine_id("opWhile")



function opNumLiteral( sequence code )
    -- { NumLiteral, val } 
    return createLiteral( code[2] )
end function
opCode[NumLiteral] = routine_id("opNumLiteral")

function opStrLiteral( sequence code )
    -- { StrLiteral, string }
    return createLiteral( code[2] )
end function
opCode[StrLiteral] = routine_id("opStrLiteral")


function opVariable( sequence code )
    -- { Variable, val }
    return javaName( code[2] )
end function
opCode[Variable] = routine_id("opVariable")

procedure callJava( sequence code )
    -- { CallProc, "java", args }
    integer at
    sequence string, token, new, tmp

    if length( code[3] ) != 1 then
        oxParseError( "java() only takes 1 argument", {} )
        return
    end if

    -- only string?    
    if code[3][1][1] != StrLiteral then
        oxParseError( "java() only takes string literals", {} )
        return
    end if

    -- get string
    string = code[3][1][2]

    -- remove the quotes
    string = string[2..length(string)-1]

    -- fix backslashed quotes
    while 1 do
        at = match( "\\\"", string )
        if at then
            string = string[1..at-1] & string[at+1..length(string)]
        else
            exit
        end if
    end while
        
    -- remove tokens from the string
    new = ""

    -- uses routine id?
    if not usesRoutineId then
        -- indirect reference?
        if match( "{class}.call_proc", string )
        or match( "{class}.call_func", string ) then
            usesRoutineId = 1
        end if
    end if


    while 1 do
        at = find( '{', string )
        if at then
            -- space follows?
            if length( string ) = at
            or  ( length( string ) > at
                  and (string[at+1] = ' ' or string[at+1] = '}') )
            then
                -- leave it alone
                new &= string[1..at]
                string = string[at+1..length(string)]
            else
                new &= string[1..at-1]
                string = string[at+1..length(string)]
                at = find( '}', string )
                if at = 0 then
                    oxParseError( "missing closing }", {} )
                    return
                end if
                token = string[1..at-1]
                string = string[at+1..length(string)]

                -- lookup token for replacement
                if equal( token, "class" ) then
                    token = class
                else
                    token = javaName( token )
                end if

                new &= token
                if oxError then
                    return
                end if
                
            end if
        else
            exit
        end if
    end while
    new &= string
    --comment( "hardcoded", {} )
    emit( 8, new, {} )

    -- was this a return statement?
    if match( "return ", ltrim( new ) ) = 1 then
        -- set position of last return
        lastReturnAt = length( out[emitTo] )
    end if
    
end procedure

procedure markRoutineId( sequence lName )
    -- convert literal name back to value
    integer at
    sequence name, jName

    -- find name of literal in list
    at = find( lName, literalName )
    name = literalList[at]
    
    -- remove quotes
    name = name[2..length(name)-1]
    -- find the routine
    for i = length( routine ) to 1 by -1 do
        -- match name?
        if equal( routine[i][Name], name ) then

            -- already on list?
            if not find( routine[i][JavaName], routine_ids ) then
                routine_ids &= {routine[i][JavaName]}
            end if

            return
        end if
    end for
    
end procedure

function callRoutine( sequence code )
    -- { CallFunc/CallProc, name, args }

    integer routineType, argCount, internal
    sequence name, string, arg, args
    object tmp

    -- embedded java call?
    if equal( code[2], "java" ) then
        -- { CallProc, "java", { StrLiteral, "file" } }
        callJava( code )
        return True
    end if

    -- embedded java_import call?
    if equal( code[2], "java_import" ) then
        -- { CallProc, "java_import", { StrLiteral, "file" } }
        string = code[3][1][2]
        string = string[2..length(string)-1]

        -- not already on import list?    
        if not find( string, jimports ) then
            jimports &= { string }
        end if
        return True
    end if

    -- decode type
    if code[1] = CallProc then
        routineType = Procedure
    else
        routineType = Function
    end if

    -- get name and args
    name = code[2]
    args = code[3]

    -- get the count of args
    argCount = length( args )

    -- check the arg count
    checkArgCount( name, argCount )

    -- builtin?               
    internal = 0
    for i = 1 to length( builtin ) by 4 do
        if equal( name, builtin[i] ) then
            internal = i
            exit
        end if
    end for

    -- build evaluated arg list
    arg = repeat( 0, argCount )
    for i = 1 to argCount do
        -- evaluate the arg
        arg[i] = eval( args[i] )
    end for

    -- builtin?  
    if internal then     
    
        -- special args for builtin?
        if equal( name, "routine_id" ) then

            -- flag routine_id is used
            usesRoutineId = 1

            -- add called routine to routine id list
            markRoutineId( arg[1] )
        
            -- add scope to args
            arg &= { sprintf( "%d", {prefixCount} ) }
        end if                             
    
        -- replace 'class.' with actual class?
        string = builtin[internal+3]
        if match( "class.", string ) = 1 then
            string = class & string[6..length(string)]
        end if                                        

        -- build call to builtin routine
        string = sprintf( string, arg )
        
    else 
        -- add to touches list
        if not find( javaName( name ), touches ) then
            touches &= { javaName( name ) }
        end if
        
        -- build the call
        string = javaName( name ) & "( "
        
        for i = 1 to argCount do
            string &= arg[i]
            if i < argCount then
                string &= ", "
            end if
        end for
        
        string &= " )"
        
    end if

    -- function or procedure?
    if routineType = Function then
        -- get a tmp to store result
        tmp = createAnon()

        -- assign the result        
        string = sprintf( "%s = %s", {tmp, string} )
        
    else       
        -- result code for procedure
        tmp = 0
    end if
    
    -- call procedure
    comment( "call %s", {name} )
    emit( 8, "%s;", {string} )

    -- free temps
    freeAnon( 1, arg )
    
    -- return result
    return tmp
    
end function

opCode[CallProc] = routine_id("callRoutine")
opCode[CallFunc] = routine_id("callRoutine")

function defRoutine( sequence code )
    -- { DefProc/DefFunc/DefType, Global/Local, name, args }
    integer isGlobal, anons, this
    sequence name, args, varName, argName, argType, returns, 
                header, priorAnons, tmp, argList, priorTouches

    -- clear touches 
    priorTouches = touches
    touches = {}

    -- function or procedure?
    if code[1] = DefProc then
        isFunction = False
        returns = "void"
    else
        isFunction = True
        returns = "Object"
    end if

    -- emit routine
    out[ToTmp] = ""
    emitTo = ToTmp

    -- decode       
    isGlobal = code[2]
    name = code[3]
    args = code[4]
    
    -- change the scope
    theScope = Routine

    -- global?    
    if isGlobal then
        -- add to global token list
        this = defineRoutine( Global, name, length( args ) )
    else
        -- add to local token list
        this = defineRoutine( Module, name, length( args ) )
    end if
    
    -- name conflict?
    if oxError then
        return True
    end if

    -- clear the lists
    argList = {}
    derefList = {}

    -- build the header
    if length( args ) then
            
        -- the args
        for i = 1 to length( args ) do
            -- get name and type
            argType = args[i][1]
            argName = args[i][2] 
        
            -- add names to routine tokens
            defineVar( Routine, argName, argType )

            -- add to arglist
            argList &= {argName}
                    
            -- add to arg list
            if i = 1 then
                if length( args ) = 1 then          
                    -- only arg
                    emit( 4, "public static %s %s ( Object %s ) {", 
                            {returns, javaName(name), javaName(argName)} )
                else           
                    -- multiple args
                    emit( 4, "public static %s %s( Object %s,", 
                            {returns, javaName(name), javaName(argName)} )
                end if
                 
            elsif i = length( args ) then
                -- last arg  
                emit( 27, "Object %s ) {", 
                        {javaName(argName)} )
            else                     
                emit( 27, "Object %s,", 
                    {javaName(argName)} )
            end if
            
        end for
    
    else
        -- no args
        emit( 4, "public static %s %s() {", {returns, javaName(name)} )
    end if

    -- build the deref list
    derefList = {}
    for i = 1 to length( argList ) do
        argName = argList[i]
        if not isNumeric( argName ) then
            derefList &= {argName}
        end if
    end for

    -- any args need to be referenced?
    if length( derefList ) then
        comment( "refcount the args", {} )
        for i = 1 to length( derefList ) do
            ref( javaName(derefList[i]) )
        end for
    end if
    
    -- save header and clear
    header = out[ToTmp]
    out[ToTmp] = ""

    -- set position of last return
    lastReturnAt = 0

    -- save temp variables state
    priorAnons = { usedAnonList, freeAnonList }
    
    -- clear temp variables
    usedAnonList = {}
    freeAnonList = {}
          
    -- parse the body of the routine
    if code[1] = DefFunc then
        parseToEnd( Function, "function" )
        
    elsif code[1] = DefType then
        parseToEnd( Type, "type" )
        
    else        
        parseToEnd( Procedure, "procedure" )
    end if          

    -- failed?
    if oxError then
        return 0
    end if

    -- create a holder for this routine
    routine[this][JavaName]   = javaName(name)
    routine[this][Returns]    = returns
    routine[this][Code]       = {}
    routine[this][Args]       = args
    routine[this][IsTouched]  = 0
    routine[this][IsRoutineId]= 0
    routine[this][Touches]    = touches
    
    -- build the routine
    emitTo = ToRoutine

    -- emit the header
    emit( 0, "%s", {header} )

    -- the variables
    if length( var[Routine] ) then
        comment( "local variables", {} )
        for i = 1 to length( var[Routine] ) do
            -- get name
            varName = var[Routine][i]
            
            -- not an arglist item?
            if not find( varName, argList ) then
                -- declare as local
                emit( 8, "Object %s = null;", {javaName(varName)} )
            end if
        end for
    end if

    -- anonymous vars        
    anons = length( usedAnonList ) + length( freeAnonList )
    if anons then
        comment( "temporary variables", {} )
        for i = 1 to anons do
            emit( 8, "Object _tmp%d = null;", {i-1} )
        end for
    end if          
          
    -- if function, allocate a variable for return value
    if isFunction then
        comment( "holds return value", {} )
        emit( 8, "Object _returnVal;", {} )
    end if

    -- emit the code
    emit( 0, "%s", {out[ToTmp]} )
              
    -- need to clean up? 
    -- this can be skipped if a return statement preceeds the end
    if lastReturnAt = 0
    or lastReturnAt != length( out[ToTmp] ) then
    
        -- function?
        if isFunction then
            comment( "error: exiting without returning value", {} )
            emit( 8, "return Eu.badExit( \"%s\" );", {name} )
    
        else
            -- deref the args
            comment( "deref the arg list", {} )
            for i = 1 to length( derefList ) do
                argName = derefList[i]
                if not isNumeric( argName ) then
                    deref( javaName(argName) )
                end if
            end for
        end if
        
    end if
    
    -- close the routine
    if code[1] = DefFunc then
        comment( "end function", {} )
        
    elsif code[1] = DefType then
        comment( "end type", {} )
        
    else
        comment( "end procedure", {} )
    end if
    
    emit( 4, "}", {} )
    emit( 0, "", {} )

    -- save the generated code
    routine[this][Code] = out[ToRoutine]
    out[ToRoutine] = {}

    -- clear values
    forgetAll( Routine )
    derefList = {}

    -- restore scope
    theScope = Module
    
    -- restore touches list
    touches = priorTouches

    -- clear function flag        
    isFunction = 0
    
    -- emit to module
    emitTo = ToModule

    -- restore temp variables state
    usedAnonList = priorAnons[1]
    freeAnonList = priorAnons[2]

    -- clear return
    lastReturnAt = 0
    
    return 0

end function
opCode[DefProc] = routine_id("defRoutine")
opCode[DefFunc] = routine_id("defRoutine")



procedure buildShadowRoutine( integer this )

    -- emit shadow code for routine_id to call
    
    sequence name, jName, args, returns, tmp

    emitTo = ToTmp
    out[ToTmp] = {}

    -- get values
    name    = routine[this][Name]
    jName   = routine[this][JavaName]
    args    = routine[this][Args]
    returns = routine[this][Returns]        

    -- create a routine for routine_id
    emit( 4, "// %s for routine_id", {name} )
    emit( 4, "public static %s _%s ( Seq s ) {", {returns, jName} )

    -- holds arg list
    tmp = ""

    -- args?
    if length( args ) then
        -- build temps to hold args
        for i = 1 to length( args ) do
            tmp &= sprintf( "_%d", {i} )
            if i != length( args ) then
                tmp &= ", "
            end if
        end for
        emit( 8, "Object %s;", {tmp} )
    
        -- assign to args
        comment( "read parms from arglist", {} )
        for i = 1 to length( args ) do
            emit( 8, "_%d = s.data[%d];", {i,i-1} )
        end for

    end if
    
    -- returns a value?
    if isFunction then
        emit( 8, "return %s( %s );", {jName,tmp} )
    else
        emit( 8, "%s( %s );", {jName,tmp} )
    end if
    emit( 4, "}", {} )
    emit( 0, "", {} )
    
end procedure

function defType( sequence code )
    -- { DefType, global flag, name, args }
    integer isGlobal, scope, at
    sequence name, args, typeof

    -- decode       
    isGlobal = code[2]
    name     = code[3]
    args     = code[4]
    
    if length( args ) != 1 then
        oxParseError( "types must have exactly one parameter", {} )
        return True
    end if

    -- look for type
    typeof = ""
    for i = 1 to length( SearchOrder ) do
        scope = SearchOrder[i]
        at = find( name, types[scope] )
        if at then
            typeof = typedefs[scope][at]
            exit
        end if
    end for
    
    -- no match?
    if length( typedefs ) = 0 then
        oxParseError( "a type is expected here", {} )
        return True
    end if
    
    -- scope of declaration?
    if isGlobal then        
        scope = Global
    else
        scope = Module
    end if                  
    
    -- add to definitions
    types[scope]     &= {name}
    typedefs[scope]  &= {typeof}

    -- create as a function
    return defRoutine( code )
    
end function
opCode[DefType] = routine_id("defType")



function opReturn( sequence code )
    -- { Return }
    sequence result, returns
    
    -- this code behaves differently, depending on whether
    -- it's called from a procedure or a function

    -- function?
    if isFunction then
    
        -- parse return value
       result = oxParse( "Expr" ) 
       if oxError then
           return 0
       end if
       
        -- evaluate
        emit( 8, "_returnVal = %s;", {eval( result )} )
        
    end if

    -- clean up           
    if length( derefList ) then
        comment( "deref the args", {} )
        for i = 1 to length( derefList ) do
            deref( javaName(derefList[i]) )
        end for
    end if

    -- return value?
    if isFunction then
        comment( "return value", {} )
        emit( 8, "return _returnVal;", {} )
    else
        comment( "return", {} )
        emit( 8, "return;", {} )
    
    end if

    -- set position of last return
    lastReturnAt = length( out[emitTo] )

    return 0
    
end function
opCode[Return] = routine_id("opReturn")


function unaryOp( sequence code )
    -- { Negate/Not, expr }
    sequence op, tmp

    -- create holder for result
    tmp = createAnon()

    if code[1] = Not then 
        comment( "not", {} )
        op = "NOT"
        
    elsif code[1] = Negate then 
        comment( "negate", {} )
        op = "NEGATE"
    end if

    -- unary op
    emit( 8, "%s = Eu.unaryOp( Eu.%s, %s );", {tmp, op, eval(code[2])} )
    
    return tmp
    
end function
opCode[Not] = routine_id("unaryOp")
opCode[Negate] = routine_id("unaryOp")

function opEnd( sequence code )
    -- { End, type }         
    sequence name
    
    if code[2] = Procedure
    or code[2] = Function then
            
        
    else
        puts( 1, "opEnd not completely implemented\n" )
        abort(0)
    end if 

    -- emit to module
    emitTo = ToModule
    
    return 0
    
end function
opCode[End] = routine_id("opEnd")

function rep( object o, integer i )
    -- "flat" version of repeat
    sequence s
    s = ""
    for j = 1 to i do
        s = s & o
    end for
    return s
end function

function opGetIndex( sequence code )
    -- { GetIndex, varName, indexes }
    -- return indexed variable  
    integer slice
    sequence name, source, target, indexes, index1, index2

    -- not a slice
    slice = 0

    -- get name of variable
    name = code[2]
    source = javaName( name )
    
    -- get indexes
    indexes = code[3]
    
    -- target
    target = createAnon()
    
    for i = 1 to length( indexes ) do
    
        -- slice?
        if indexes[i][1] = Slice then
            -- { Slice, low, high }
            slice = 1

            -- evaluate the indexes
            index1 = eval( indexes[i][2] )
            index2 = eval( indexes[i][3] )

            -- get the element
            comment( "slice from %s%s", {code[2], rep("[n]",i-1)})
            emit( 8, "%s = Eu.getSlice( %s, %s, %s );", 
                {target, source, index1, index2} )
            
        else
            -- evaluate the index
            index1 = eval( indexes[i] )
            
            -- get the element
            comment( "indexed fetch from %s%s", {code[2], rep("[n]",i)})
            emit( 8, "%s = Eu.getItem( %s, %s );", {target, source, index1} )
            
            -- change source
            source = target
            
        end if
                
            
    end for

    -- refcount the object
    ref( target )
    
    -- release the index
    if slice then           
        -- both vars
        freeAnon( 0, {index1, index2})
    else            
        -- only one var used
        freeAnon( 0, {index1})
    
    end if
    
    return target

end function
opCode[GetIndex] = routine_id("opGetIndex")

function opExit( sequence code )
    -- { Exit }

    comment( "exit", {} )
    emit( 8, "break;", {} )
    
    -- write the module code
    return True
    
end function
opCode[Exit] = routine_id("opExit")


global procedure parseFile( sequence fileName )

    -- parse the file
    -- shared by EJ.EX
    integer finished
    object result

    -- set the file
    oxLoadFile( fileName )

    -- until end of file
    while 1 do
    
        -- end of file?
        result = oxTokenPeek()
        if equal( result[1], "<EOF>") then
            -- all done
            exit
        end if
        
        -- parse
        result = oxParse( "Command" )
        if oxError then
            -- quit
            exit
        end if
        
        -- execute
        result = eval( result )
        
    end while

    -- close the file
    oxUnloadFile()
    
end procedure

function opInclude( sequence code )
    -- { Include, filename }
    -- include file             
    integer at, handle, finished
    sequence filename, currentPath, includePath, fullPath, usePath, prior
    object result

    -- called in routine?
    if theScope = Routine then                                   
        -- oxParseError( "expected to see possibly 'end' not 'include'", {} )
        oxParseError( "can't use 'include' inside routines", {} )
        return True
    end if

    -- get the file name from the input stream
    result = oxLineGet()

    -- find the include statement and remove it
    at = match( "include", result )
    result = result[at+7..length(result)]
    
    -- trim on both sides
    result = rtrim( result )
    result = ltrim( result )
    
    -- ends with space?
    at = find( ' ', result )
    if at then
        result = result[1..at-1]
    end if
    
    -- set as filename
    filename = result
    
    -- no name?
    if equal( filename, "" ) then
        oxParseError( "file name is missing", {} )
        return True
    end if

    -- quoted?
    if filename[1] = '\"' then
        oxParseError( "do not put quotes around an include file name", {} )
        return True
    end if

    -- get full pathname of current directory
    currentPath = current_dir() & "\\" & filename

    -- get the path for the include file
    result = getenv("EUDIR")
    
    -- not defined?
    if equal( result, -1) then
        -- default to current drive
        includePath = "\\EUPHORIA\\include\\" & filename
    else                    
        -- use value returned from EUDIR
        includePath = result
    end if

    -- look in current directory
    result = dir( currentPath )
    if equal( result, -1 ) then
        -- look in include directory
        result = dir( includePath )
        -- store as potentially good path
        usePath = includePath
    else                  
        -- store as good path
        usePath = currentPath         
    end if
    
    -- found the file?
    if equal( result, -1 ) then
        -- mark handle as bad
        handle = -1
    else
        -- get name
        filename = result[1][D_NAME]
        
        -- already loaded?
        if find( filename, included ) then
            -- skip
            return True
        end if
        
        -- try to open the file
        handle = open( usePath, "r" )
        
    end if
    
    -- failed to find or open?
    if handle = -1 then
        oxParseError( "can't open %s or %s", 
            {currentPath, includePath} )
        return True
    else               
        --  close the file
        close( handle )
    end if         

    -- add to the included list
    included &= {filename}

    -- save current information
    prior = { 
        prefix,             -- the prefixes
        token[Module],      -- the tokens
        token_args[Module], -- argument counts
        token_type[Module], -- datatype
        var[Module],        -- variables
        types[Module],      -- token types
        typedefs[Module] }  -- type conversions
    
    -- clear module information
    forgetAll( Module )
        
    -- change prefixes
    prefixCount += 2
    prefix = { "_0", 
               sprintf( "_%d", prefixCount ),
               sprintf( "_%d", prefixCount+1 ) }    
    
    -- increase size of module variables
    moduleVars &= repeat( {}, 2 )
    
    -- parse the file
    parseFile( usePath )

    -- restore module settings
    prefix              = prior[1]
    token[Module]       = prior[2]
    token_args[Module]  = prior[3]
    token_type[Module]  = prior[4]
    var[Module]         = prior[5]
    types[Module]       = prior[6]
    typedefs[Module]    = prior[7]
    
    return True
    
end function
opCode[Include] = routine_id("opInclude")


function opGlobalFlag( sequence code )
    -- { GlobalFlag }
    sequence nextToken

    -- called in routine?
    if theScope = Routine then                                   
        -- oxParseError( "expected to see possibly 'end' not 'global'", {} )
        oxParseError( "can't use 'global' inside routines", {} )
        return True
    end if

    -- check following token
    nextToken = oxTokenPeek()
    if find( nextToken[1], {"constant", "procedure", "type", "function"} )
    or equal( nextToken[2], "TYPE") then
        -- set the global flag
        globalFlag = True
    else
        -- error
        oxParseError( "'global' must be followed by:\n" &
                    "<a type>, 'constant', 'procedure', 'type' or 'function'", {} )
        return True
    end if

    return True
        
end function
opCode[GlobalFlag] = routine_id( "opGlobalFlag" )

procedure findTouched( sequence list )
    -- flag all the routines that are on the list as touched

    -- look at each routine
    for i = 1 to length( routine ) do
        -- on list and not already touched?
        if not routine[i][IsTouched]
        and find( routine[i][JavaName], list ) then
            -- touch
            routine[i][IsTouched] = 1
            -- touch all routines this touches
            findTouched( routine[i][Touches] )
        end if
    end for
end procedure


procedure buildRoutineIdCode()
    
    -- build support for routine_id in out[ToTmp]
    integer count, this
    sequence format, jName

    -- mark all routines that are referenced by routine_id
    count = 0
    for i = 1 to length( routine ) do
        jName = routine[i][JavaName]
        if find( jName, routine_ids ) then
            -- flag
            routine[i][IsRoutineId] = 1
            count += 1                
            
            -- mark as touched
            findTouched( {jName} )
        end if
    end for
    
    emitTo = ToTmp
    out[emitTo] = ""
    
    -- build the structures
    emit( 0, "", {} )
    emit( 4, "// support for routine_id", {} )
    emit( 4, "static %s instanceOf_%s = new %s();", {class,class,class} )
    emit( 4, "static Class class_%s = instanceOf_%s.getClass();", {class,class} )
    emit( 4, "static Object[] argList = { new Seq() };", {} )
    emit( 4, "static Class[] argListClass = { argList[0].getClass() };", {} )

    -- method for getting the routine id    
    emit( 0, "", {} )
    emit( 4, "public static java.lang.reflect.Method getMethodId( String methodName ) {", {} )
    emit( 4, "    // get the methods", {} )
    emit( 4, "    try { return class_%s.getMethod( methodName, argListClass ); }", {class} )
    emit( 4, "    catch (Exception e) {", {} )
    emit( 4, "        Eu.fatal(\"getMethodId: exception thrown for \" + methodName);", {} )
    emit( 4, "        return null;", {} )
    emit( 4, "    }", {} )
    emit( 4, "}", {} )


    -- get the id's and store in an array                     
    emit( 0, "", {} )
    emit( 4, "// store methods", {} )
    emit( 4, "static java.lang.reflect.Method[] routine_method = {", {} )
    this = 0
    for i = 1 to length( routine ) do     
        if routine[i][IsRoutineId] then
            -- format      
            format = "%s.getMethodId(\"_%s\")"
            
            -- add comma?
            this += 1
            if this != count then
                format &= ","
            end if
        
            -- emit code
            emit( 8, format, {class,routine[i][JavaName]} )
        end if
    end for
    emit( 4, "};", {} )

    -- get the method names and store in an array                     
    emit( 0, "", {} )
    emit( 4, "// euphoria names", {} )
    emit( 4, "static String[] routine_name = {", {} )
    this = 0
    for i = 1 to length( routine ) do
        if routine[i][IsRoutineId] then
            -- format      
            format = "\"%s\""
            
            -- add comma?   
            this += 1
            if this != count then
                format &= ","
            end if
        
            -- emit code
            emit( 8, format, {routine[i][Name]} )
            
        end if
    end for
    emit( 4, "};", {} )

    -- get the method function flag and store in an array                     
    emit( 0, "", {} )
    emit( 4, "// method args", {} )
    emit( 4, "static boolean[] routine_isFunc = {", {} )
    this = 0
    for i = 1 to length( routine ) do
        if routine[i][IsRoutineId] then

            if equal( routine[i][Returns], "void" ) then
                format = "false"
            else
                format = "true"
            end if
            
            -- add comma?
            this += 1
            if this != count then
                format &= ","
            end if
    
            -- emit code
            emit( 8, format, {} )
            
        end if
            
    end for
    emit( 4, "};", {} )


    -- get the method arg count and store in an array                     
    emit( 0, "", {} )
    emit( 4, "// method args", {} )
    emit( 4, "static int[] routine_args = {", {} )
    
    this = 0
    for i = 1 to length( routine ) do
        if routine[i][IsRoutineId] then
        
            -- format      
            format = "%d"
            
            -- add comma?
            this += 1
            if this != count then
                format &= ","
            end if
        
            -- emit code
            emit( 8, format, {length( routine[i][Args] )} )
            
        end if
        
    end for
    emit( 4, "};", {} )

    -- get the method scope and store in an array                     
    emit( 0, "", {} )
    emit( 4, "// method scope", {} )
    emit( 4, "static int[] routine_scope = {", {} )
    
    this = 0
    for i = 1 to length( routine ) do
        if routine[i][IsRoutineId] then
            -- format      
            format = "%d"
            
            -- add comma?
            this += 1
            if this != count then
                format &= ","
            end if
        
            -- emit code
            emit( 8, format, {routine[i][Scope]} )
        
        end if
    end for
    emit( 4, "};", {} )

    -- call_proc    
    emit( 0, "", {} )
    emit( 4, "// call_proc", {} )
    emit( 4, "public static void call_proc( Object n, Object s ) {", {} )
    
    emit( 0, "", {} )
    emit( 4, "    // args are sequence?", {} )
    emit( 4, "    if (!(s instanceof Seq)) {", {} )
    emit( 4, "        Eu.fatal( \"argument list must be a sequence\");", {} )
    emit( 4, "    }", {} )
    
    emit( 0, "", {} )
    emit( 4, "    // id not numeric?", {} )
    emit( 4, "    if (n instanceof Seq) {", {} )
    emit( 4, "        Eu.fatal( \"call_proc: an integer was expected, not a sequence\");", {} )
    emit( 4, "    }", {} )
    
    emit( 0, "", {} )
    emit( 4, "    // valid id?", {} )
    emit( 4, "    int id = (int)(Eu.toNumber( n ));", {} )
    emit( 4, "    if (id < 0 || id > %d || id != Math.floor(id)) {", {length(routine_ids)-1})
    emit( 4, "        Eu.fatal( \"invalid routine id\");", {} )
    emit( 4, "    }", {} )
    
    emit( 0, "", {} )
    emit( 4, "    // check arg count", {} )
    emit( 4, "    if (((Seq)s).length != routine_args[id]) {", {})
    emit( 4, "        Eu.fatal( \"call to \" + routine_name[id] + ", {} )
    emit( 4, "            \" via routine_id should pass \" + routine_args[id] +", {} )
    emit( 4, "            \" arguments, not \" + ((Seq)s).length);", {} )
    emit( 4, "    }", {} )    
    
    emit( 0, "", {} )
    emit( 4, "    // set the parameter", {} )
    emit( 4, "    argList[0] = s;", {} )

    emit( 0, "", {} )
    emit( 4, "    // function?", {} )
    emit( 4, "    if (routine_isFunc[id]) {", {})
    emit( 4, "        // call the static function", {} )
    emit( 4, "        try { Object result = %s.routine_method[id].invoke( null, argList ); }", {class} )
    emit( 4, "        catch (Exception e) {", {} )
    emit( 4, "            Eu.fatal( \"call_proc: illegal exception, calling \" + %s.routine_name[id] );", {class} )
    emit( 4, "        }", {} )
    emit( 4, "    } else {", {} )    
    emit( 4, "        // call the static procedure", {} )
    emit( 4, "        try { %s.routine_method[id].invoke( null, argList ); }", {class} )
    emit( 4, "        catch (Exception e) {", {} )
    emit( 4, "            Eu.fatal( \"call_proc: illegal exception, calling \" + %s.routine_name[id] );", {class} )
    emit( 4, "         }", {} )
    emit( 4, "    }", {} )
    emit( 4, "}", {} )    
                
    -- call_func
    emit( 0, "", {} )
    emit( 4, "// call_func", {} )
    emit( 4, "public static Object call_func( Object n, Object s ) {", {} )
    
    emit( 0, "", {} )
    emit( 4, "    // args are sequence?", {} )
    emit( 4, "    if (!(s instanceof Seq)) {", {} )
    emit( 4, "        Eu.fatal( \"argument list must be a sequence\");", {} )
    emit( 4, "    }", {} )
    
    emit( 0, "", {} )
    emit( 4, "    // id not numeric?", {} )
    emit( 4, "    if (n instanceof Seq) {", {} )
    emit( 4, "        Eu.fatal( \"call_proc: an integer was expected, not a sequence\");", {} )
    emit( 4, "    }", {} )
    
    emit( 0, "", {} )
    emit( 4, "    // valid id?", {} )
    emit( 4, "    int id = (int)(Eu.toNumber( n ));", {} )
    emit( 4, "    if (id < 0 || id > %d || id != Math.floor(id)) {", {length(routine_ids)-1})
    emit( 4, "        Eu.fatal( \"invalid routine id\");", {} )
    emit( 4, "    }", {} )

    emit( 0, "", {} )
    emit( 4, "    // check arg count", {} )
    emit( 4, "    if (((Seq)s).length != routine_args[id]) {", {})
    emit( 4, "        Eu.fatal( \"call to \" + routine_name[id] + ", {} )
    emit( 4, "            \" via routine_id should pass \" + routine_args[id] +", {} )
    emit( 4, "            \" arguments, not \" + ((Seq)s).length);", {} )
    emit( 4, "    }", {} )    

    emit( 0, "", {} )
    emit( 4, "    // function?", {} )
    emit( 4, "    if (!routine_isFunc[id]) {", {})
    emit( 4, "        Eu.fatal( routine_name[id] + \" does not return a value\" );", {} )
    emit( 4, "    }", {} )    
    
    emit( 0, "", {} )
    emit( 4, "    // set the parameter", {} )
    emit( 4, "    argList[0] = s;", {} )
    
    emit( 0, "", {} )
    emit( 4, "    // invoke the static method", {} )
    emit( 4, "    try { return %s.routine_method[id].invoke( null, argList ); }", {class} )
    emit( 4, "    catch (Exception e) {", {} )
    emit( 4, "        Eu.fatal( \"call_proc: illegal exception, calling \" + %s.routine_name[id] );", {class} )
    emit( 4, "        return null;", {} )
    emit( 4, "     }", {} )
    emit( 4, "}", {} )    

    emit( 0, "", {} )
    emit( 4, "// find a routine", {} )
    emit( 4, "public static Object routine_id( Object name, int scope ) {", {} )
    emit( 4, "    // check the type", {} )
    emit( 4, "    if (!(name instanceof Seq)) {", {} )
    emit( 4, "        return Eu.create( -1 );", {} )
    emit( 4, "    }", {} )
    emit( 0, "", {} )
    emit( 4, "    // convert to string", {} )
    emit( 4, "    String s = (Eu.seqToStringBuffer((Seq)name)).toString();", {} )
    emit( 4, "    for( int i = 0; i < %d; i++ ) {", {length(routine_ids)} )
    emit( 4, "        if (routine_scope[i] == scope || routine_scope[i] == 0) {", {} )
    emit( 4, "            if (s.equals( routine_name[i] )) {", {} )
    emit( 4, "                return Eu.create( i );", {} )
    emit( 4, "            }", {} )
    emit( 4, "        }", {} )
    emit( 4, "    }", {} )
    emit( 4, "    return Eu.create( -1 );", {} )
    emit( 0, "}", {} )
                                                                     
end procedure


global procedure writeFiles( sequence fname )
    -- generate the output files
    integer handle, anons, at
    sequence name, val, this

    -- output file
    handle = open( "out", "w" )

    puts( handle, "// generated with EuJava beta\n" )
    puts( handle, "// (c) 2001 David Cuny\n" )
    puts( handle, "// email: dcuny@lanset.com\n" )
    puts( handle, "//\n" )
    puts( handle, "// Thanks to Robert Craig of RDS\n" )
    puts( handle, "// Euphoria is (c) 2001 Rapid Development Software\n" )
    puts( handle, "// http://www.rapideuphoria.com\n" )

    -- imports: built from java_import() calls
    for i = 1 to length( jimports ) do
        printf( handle, "import %s;\n", {jimports[i]} )
    end for
    
    puts( handle, "\n" )
    printf( handle, "class %s {\n", {class} )
    puts( handle, "\n" )
    
    -- the literals
    if length( literalList ) then
        puts( handle, "\n" )
        puts( handle, tab(4) & "// literals\n" ) 
        for i = 1 to length( literalList ) do
            val = literalList[i]
            name = literalName[i]
            if val[1] = '\"' then
                printf( handle, 
                    "    static Object %s = (Object) new Seq( %s );\n", 
                    {name, val})
            else
                printf( handle, 
                    "    static Object %s = Eu.create( %s );\n", 
                    {name,val})
            end if
        end for
        
    end if

    -- if flag                      
    puts( handle, "\n" )
    puts( handle, tab(4) & "// test flag\n" )
    puts( handle, tab(4) & "static boolean _test;\n" )

    -- anonymous vars        
    anons = length( usedAnonList ) + length( freeAnonList )
    if anons then
        puts( handle, "\n" )                      
        puts( handle, tab(4) & "// temporary variables\n" )
        for i = 1 to anons do
            printf( handle, tab(4) & "static Object _tmp%d;\n", {i-1} )
        end for
    end if

    -- module level variables      
    if length( moduleVars ) then
        puts( handle, "\n\n" )
        puts( handle, "    // module variables\n" )
        for i = 1 to length( moduleVars ) do
            this = moduleVars[i]
            for j = 1 to length( this ) do         
                printf( handle, "    static Object %s = null;\n", {this[j]} )
            end for
        end for
    end if

    -- globals
    if length( var[Global] ) then
        puts( handle, "\n\n" )
        puts( handle, "    // global variables\n" )
        for i = 1 to length( var[Global] ) do
            name = var[Global][i]
            printf( handle, 
                tab(4) & "static Object %s = null;\n", 
                {javaName(name)} )
        end for
    end if

    -- routine id support
    if usesRoutineId then
        buildRoutineIdCode()
        puts( handle, out[ToTmp] & "\n\n" )
        out[ToTmp] = ""
    end if

    -- procedures and functions
    if length( routine ) then
        puts( handle, "\n" )
        puts( handle, tab(4) & "// procedures and functions\n" )
        
        -- find out what routines were touched
        findTouched( touches )

        -- look at each routine
        for i = 1 to length( routine ) do
            -- marked as called?
            if routine[i][IsTouched] then
                -- emit
                puts( handle, routine[i][Code] )
                
                -- shadow routine?
                if routine[i][IsRoutineId] then
                    buildShadowRoutine( i )
                    puts( handle, out[ToTmp] )
                end if
            end if
        end for
        
    end if

    -- inline code
    if length( out[ToModule] ) then
        puts( handle, "\n" )
        puts( handle, tab(4) & "// inline code\n" )
        puts( handle, "    static {\n" )
        puts( handle, out[ToModule] & "\n" ) 
        puts( handle, tab(4) & "}\n" )
    end if

    -- main
    puts( handle, "\n" )
    puts( handle, tab(4) & "// main\n" )
    puts( handle, tab(4) & "public static void main(String[] args) {}\n" )
    puts( handle, "}\n" )
    
    -- close the file
    close( handle )
    
    -- rename the file
    system( "del \"" & class & ".java\"", 2 )
    system( "rename out \"" & class & ".java\"", 2 )
    
end procedure

global procedure mainModule( sequence fname )
end procedure

